class Singleton(type):
    """Metaclass that makes a class into a singleton."""

    def __new__(cls, name, bases, cls_dict):
        singleton_cls = super().__new__(cls, name, bases, cls_dict)

        def clean():
            if hasattr(singleton_cls, "instance"):
                delattr(singleton_cls, "instance")

        setattr(singleton_cls, "clean", clean)
        return singleton_cls

    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, "instance"):
            cls.instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.instance


class SingletonNotInstantiatedException(Exception):
    """Raised when a singelton is accessed without being initialized."""



import logging
import os
from contextlib import contextmanager
from typing import Optional

from rdflib import Graph
from rdflib.namespace import NamespaceManager
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from buildingmotif.building_motif.singleton import (
    Singleton,
    SingletonNotInstantiatedException,
)
from buildingmotif.database.graph_connection import GraphConnection
from buildingmotif.database.table_connection import TableConnection
from buildingmotif.database.tables import Base as BuildingMOTIFBase
from buildingmotif.database.utils import (
    _custom_json_deserializer,
    _custom_json_serializer,
)
from buildingmotif.namespaces import bind_prefixes


class BuildingMOTIF(metaclass=Singleton):
    """Manages BuildingMOTIF data classes."""

    def __init__(
        self,
        db_uri: str,
        shacl_engine: Optional[str] = "pyshacl",
        log_level=logging.WARNING,
    ) -> None:
        """Class constructor.

        :param db_uri: database URI
        :type db_uri: str
        :param shacl_engine: the name of the engine to use for validation: "pyshacl" or "topquadrant". Using topquadrant
            requires Java to be installed on this machine, and the "topquadrant" feature on BuildingMOTIF,
            defaults to "pyshacl"
        :type shacl_engine: str, optional
        :param log_level: logging level of detail
        :type log_level: int
        :default log_level: INFO
        """
        self.db_uri = db_uri
        self.shacl_engine = shacl_engine or "pyshacl"
        self.engine = create_engine(
            db_uri,
            echo=False,
            json_serializer=_custom_json_serializer,
            json_deserializer=_custom_json_deserializer,
        )
        self.session_factory = sessionmaker(bind=self.engine, autoflush=True)
        self.Session = scoped_session(self.session_factory)

        self.setup_logging(log_level)

        # setup tables automatically if using a in-memory sqlite database
        if self._is_in_memory_sqlite():
            self.setup_tables()

        self.table_connection = TableConnection(self.engine, self)
        self.graph_connection = GraphConnection(
            BuildingMotifEngine(self.engine, self.Session)
        )

        g = Graph()
        bind_prefixes(g)
        self.template_ns_mgr: NamespaceManager = NamespaceManager(g)

    @property
    def session(self):
        return self.Session()

    def setup_tables(self):
        """Creates all tables in the underlying database."""
        BuildingMOTIFBase.metadata.create_all(self.engine)

    def _is_in_memory_sqlite(self) -> bool:
        """Returns true if the BuildingMOTIF instance uses an in-memory SQLite
        database.
        """
        if self.engine.dialect.name != "sqlite":
            return False
        # get the 'filename' of the database; if this is empty, the db is in-memory
        raw_conn = self.engine.raw_connection()
        filename = (
            raw_conn.cursor()
            .execute("select file from pragma_database_list where name='main';", ())
            .fetchone()
        )
        # length is 0 if the db is in-memory
        return not len(filename[0])

    def setup_logging(self, log_level):
        """Create log file with DEBUG level and stdout handler with specified
        logging level.

        :param log_level: logging level of detail
        :type log_level: int
        """
        root_logger = logging.getLogger()
        root_logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s | %(name)s |  %(levelname)s: %(message)s"
        )

        log_file_handler = logging.FileHandler(
            os.path.join(os.getcwd(), "BuildingMOTIF.log"), mode="w"
        )
        log_file_handler.setLevel(logging.DEBUG)
        log_file_handler.setFormatter(formatter)

        engine_logger = logging.getLogger("sqlalchemy.engine")
        pool_logger = logging.getLogger("sqlalchemy.pool")

        engine_logger.setLevel(logging.WARN)
        pool_logger.setLevel(logging.WARN)

        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(log_level)
        stream_handler.setFormatter(formatter)

        root_logger.addHandler(log_file_handler)
        root_logger.addHandler(stream_handler)

    def close(self) -> None:
        """Close session and engine."""
        self.session.close()
        self.engine.dispose()


def get_building_motif() -> "BuildingMOTIF":
    """Returns singleton instance of BuildingMOTIF.

    Requires that BuildingMOTIF has been instantiated before, otherwise raises
    an exception.

    :raises SingletonNotInstantiatedException: if buildingmotif hasn't been
        instantiated
    :return: singleton instance of buildingmotif
    :rtype: BuildingMOTIF
    """
    if hasattr(BuildingMOTIF, "instance"):
        return BuildingMOTIF.instance  # type: ignore
    raise SingletonNotInstantiatedException


class BuildingMotifEngine:
    """BuildingMotifEngine is a class that wraps a SQLAlchemy Engine and
    Session.

    This enables the use of sessioned transactions in rdflib-sqlalchemy.
    If we are experiencing weird graph database issues this may be the cause.
    """

    def __init__(self, engine, Session) -> None:
        self.engine = engine
        self.Session = Session

    # begin and connect attributes are queried from the wrapped session.

    @contextmanager
    def begin(self):
        yield self.Session()

    @contextmanager
    def connect(self):
        yield self.Session()

    def __getattr__(self, attr):
        # When an attribute is requested, see if we have overriden it
        # If we have not return the attr of the wrapped engine
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.engine, attr)



from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, List, Optional

import rdflib
import rdflib.query
import rfc3987

from buildingmotif import get_building_motif
from buildingmotif.dataclasses.shape_collection import ShapeCollection
from buildingmotif.dataclasses.validation import ValidationContext
from buildingmotif.utils import Triple, copy_graph, shacl_inference, skolemize_shapes

if TYPE_CHECKING:
    from buildingmotif import BuildingMOTIF
    from buildingmotif.dataclasses.compiled_model import CompiledModel


def _validate_uri(uri: str):
    parsed = rfc3987.parse(uri)
    if not parsed["scheme"]:
        raise ValueError(
            f"{uri} does not look like a valid URI, trying to serialize this will break."
        )


@dataclass
class Model:
    """This class mirrors :py:class:`database.tables.DBModel`."""

    _id: int
    _name: str
    _description: str
    _graph: rdflib.Graph
    _bm: "BuildingMOTIF"
    _manifest_id: int

    @classmethod
    def create(cls, name: str, description: str = "") -> "Model":
        """Create a new model.

        :param name: new model name
        :type name: str
        :param description: new model description
        :type description: str
        :return: new model
        :rtype: Model
        """
        _validate_uri(name)
        g = rdflib.Graph()
        g.add((rdflib.URIRef(name), rdflib.RDF.type, rdflib.OWL.Ontology))
        if description:
            g.add(
                (rdflib.URIRef(name), rdflib.RDFS.comment, rdflib.Literal(description))
            )
        return cls.from_graph(g)

    @classmethod
    def from_graph(cls, graph: rdflib.Graph) -> "Model":
        """Create a new model from a graph. The name of the model is taken from the
        ontology declaration in the graph (subject of rdf:type owl:Ontology triple).
        The description of the model can be set through an RDFS comment on the ontology

        :param graph: graph to create model from
        :type graph: rdflib.Graph
        :return: new model
        :rtype: Model
        """
        bm = get_building_motif()

        name = graph.value(predicate=rdflib.RDF.type, object=rdflib.OWL.Ontology)
        if name is None:
            raise ValueError("Graph does not contain an ontology declaration")
        _validate_uri(name)

        # the 'description' is the rdfs:comment of the ontology
        description = graph.value(name, rdflib.RDFS.comment)
        description = str(description) if description is not None else ""

        db_model = bm.table_connection.create_db_model(name, description)

        graph = bm.graph_connection.create_graph(db_model.graph_id, graph)

        # below, we normalize the name to a string so it matches the database type
        return cls(
            _id=db_model.id,
            _name=str(db_model.name),
            _description=db_model.description,
            _graph=graph,
            _bm=bm,
            _manifest_id=db_model.manifest_id,
        )

    @classmethod
    def from_file(cls, url_or_path: str) -> "Model":
        """Create a new model from a file.

        :param url_or_path: url or path to file
        :type url_or_path: str
        :return: new model
        :rtype: Model
        """
        graph = rdflib.Graph()
        # if guess_format doesn't match anything, it will return None,
        # which tells graph.parse to guess 'turtle'

        # if graph parsing fails, it will raise an exception
        graph.parse(url_or_path, format=rdflib.util.guess_format(url_or_path))
        return cls.from_graph(graph)

    @classmethod
    def load(cls, id: Optional[int] = None, name: Optional[str] = None) -> "Model":
        """Get model from database by id or name.

        :param id: model id, defaults to None
        :type id: Optional[int], optional
        :param name: model name, defaults to None
        :type name: Optional[str], optional
        :raises Exception: if neither id nor name provided
        :return: model
        :rtype: Model
        """
        bm = get_building_motif()
        if id is not None:
            db_model = bm.table_connection.get_db_model(id)
        elif name is not None:
            db_model = bm.table_connection.get_db_model_by_name(name)
        else:
            raise Exception("Neither id nor name provided to load Model")
        graph = bm.graph_connection.get_graph(db_model.graph_id)

        return cls(
            _id=db_model.id,
            _name=db_model.name,
            _description=db_model.description,
            _graph=graph,
            _bm=bm,
            _manifest_id=db_model.manifest_id,
        )

    @property
    def id(self) -> Optional[int]:
        return self._id

    @id.setter
    def id(self, new_id):
        raise AttributeError("Cannot modify db id")

    @cached_property
    def graph(self) -> rdflib.Graph:
        return self._graph

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._bm.table_connection.update_db_model_name(self._id, new_name)
        self._name = new_name

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, new_description: str):
        self._bm.table_connection.update_db_model_description(self._id, new_description)
        self._description = new_description

    def add_triples(self, *triples: Triple) -> None:
        """Add the given triples to the model.

        :param triples: a sequence of triples to add to the graph
        :type triples: Triple
        """
        for triple in triples:
            self.graph.add(triple)

    def add_graph(self, graph: rdflib.Graph) -> None:
        """Add the given graph to the model.

        :param graph: the graph to add to the model
        :type graph: rdflib.Graph
        """
        self.graph += graph

    def validate(
        self,
        shape_collections: Optional[List[ShapeCollection]] = None,
        error_on_missing_imports: bool = True,
        shacl_engine: Optional[str] = None,
    ) -> "ValidationContext":
        """Validates this model against the given list of ShapeCollections.
        If no list is provided, the model will be validated against the model's "manifest".
        If a list of shape collections is provided, the manifest will *not* be automatically
        included in the set of shape collections.

        Loads all of the ShapeCollections into a single graph.

        :param shape_collections: a list of ShapeCollections against which the
            graph should be validated. If an empty list or None is provided, the
            model will be validated against the model's manifest.
        :type shape_collections: List[ShapeCollection]
        :param error_on_missing_imports: if True, raises an error if any of the dependency
            ontologies are missing (i.e. they need to be loaded into BuildingMOTIF), defaults
            to True
        :type error_on_missing_imports: bool, optional
        :return: An object containing useful properties/methods to deal with
            the validation results
        :param shacl_engine: the SHACL engine to use for validation, defaults to whatever
            is set in the BuildingMOTIF object
        :type shacl_engine: str, optional

        :rtype: ValidationContext
        """
        compiled_model = self.compile(shape_collections or [self.get_manifest()])
        return compiled_model.validate(error_on_missing_imports)

    def compile(
        self, shape_collections: Optional[List["ShapeCollection"]] = None
    ) -> "CompiledModel":
        """Compile the graph of a model against a set of ShapeCollections.

        :param shape_collections: list of ShapeCollections to compile the model
            against. Defaults to the model's manifest.
        :type shape_collections: List[ShapeCollection], optional
        :param shacl_engine: the SHACL engine to use for validation, defaults to whatever
            is set in the BuildingMOTIF object
        :type shacl_engine: str, optional
        :return: copy of model's graph that has been compiled against the
            ShapeCollections
        :rtype: Graph
        """
        from buildingmotif.dataclasses.compiled_model import CompiledModel

        ontology_graph = rdflib.Graph()
        if shape_collections is None:
            shape_collections = [self.get_manifest()]
        for shape_collection in shape_collections:
            ontology_graph += shape_collection.graph

        ontology_graph = skolemize_shapes(ontology_graph)

        model_graph = copy_graph(self.graph).skolemize()

        compiled_graph = shacl_inference(
            model_graph, ontology_graph, engine=self._bm.shacl_engine
        )
        return CompiledModel(self, shape_collections, compiled_graph)

    def get_manifest(self) -> ShapeCollection:
        """Get ShapeCollection from model.

        :return: model's shape collection
        :rtype: ShapeCollection
        """
        return ShapeCollection.load(self._manifest_id)

    def update_manifest(self, manifest: ShapeCollection):
        """Updates the manifest for this model by adding in the contents
        of the shape graph inside the provided SHapeCollection

        :param manifest: the ShapeCollection containing additional shapes against which to validate this model
        :type manifest: ShapeCollection
        """
        self.get_manifest().graph += manifest.graph



from abc import ABC
from dataclasses import dataclass
from typing import Any, Optional, Type, Union

from rdflib import URIRef


@dataclass(frozen=True)
class Token(ABC):
    value: Any


# Token is a union of the different types of tokens
TokenOrConstructor = Union[Token, Type[Token]]


def ensure_token(token_or_constructor: TokenOrConstructor, value):
    """Ensure a value is a token or constructs one from a given value."""
    if isinstance(token_or_constructor, type):
        return token_or_constructor(value)
    return token_or_constructor


@dataclass(frozen=True)
class Identifier(Token):
    """An identifier token. Contains a string."""

    value: str


@dataclass(frozen=True)
class Constant(Token):
    """A constant token. Contains a URI, probably some sort of Class"""

    value: URIRef


@dataclass(frozen=True)
class Delimiter(Token):
    """A delimiter token."""

    value: str


@dataclass(frozen=True)
class Null(Token):
    """A null token."""

    value: None = None


@dataclass(frozen=True)
class TokenResult:
    """A token result. Contains a token, the type of the token, the length of the token, and a possible error."""

    value: Optional[str]
    token: Token
    length: int
    error: Optional[str] = None
    id: Optional[str] = None

    def __eq__(self, other):
        """
        Compare two token results on every
        field except for the error field.
        """
        if not isinstance(other, TokenResult):
            return False
        return (
            self.value == other.value
            and self.token == other.token
            and self.length == other.length
        )


# null token result
ErrorTokenResult = TokenResult(None, Null(), 0, None)



import logging

import flask
from flask import Blueprint, jsonify, request
from flask_api import status

from buildingmotif.api.serializers.parser import deserialize
from buildingmotif.label_parsing.parser import parse

log = logging.getLogger()

blueprint = Blueprint("parsers", __name__)


@blueprint.route("", methods=(["POST"]))
def evaluate() -> flask.Response:
    raw_data = request.json

    my_parser = deserialize(raw_data.get("parsers"))
    point_labels = raw_data.get("point_labels")

    return (
        jsonify([parse(my_parser, point_label) for point_label in point_labels]),
        status.HTTP_200_OK,
    )



from typing import List, Union

from typing_extensions import TypedDict

from buildingmotif.database.tables import DBModel

ModelDict = TypedDict(
    "ModelDict",
    {
        "id": int,
        "name": str,
        "description": str,
        "graph_id": str,
    },
)


def serialize(
    param: Union[DBModel, List[DBModel]]
) -> Union[ModelDict, List[ModelDict]]:
    """Serialize one or more models into a TypedDict.

    :param param: one model or a list of models
    :type param: Union[DBModel, List[DBModel]]
    :return: one JSON per serialized model
    :rtype: Union[ModelDict, List[ModelDict]]
    """
    if isinstance(param, DBModel):
        return _serialize(param)

    elif isinstance(param, list):
        return [_serialize(x) for x in param]

    raise ValueError("invalid input. Must be a DBModel or list of DBModels")


def _serialize(model: DBModel) -> ModelDict:
    """Serialize a model into a TypedDict.

    :param model: model
    :type model: DBModel
    :return: serialized model
    :rtype: ModelDict
    """
    return {
        "id": model.id,
        "name": model.name,
        "description": model.description,
        "graph_id": model.graph_id,
    }



import logging
from functools import cached_property
from os import PathLike
from typing import List, Optional

from buildingmotif.ingresses.base import Record, RecordIngressHandler

try:
    from openpyxl import load_workbook
except ImportError:
    logging.critical(
        "Install the 'xlsx-ingress' module, e.g. 'pip install buildingmotif[xlsx-ingress]'"
    )


class XLSXIngress(RecordIngressHandler):
    """Reads sheets from a XLSX file and exposes them as records. The 'rtype'
    field of each Record gives the name of the sheet.
    """

    def __init__(self, filename: PathLike, limit: Optional[int] = -1):
        """
        Path to the .xlsx file to be ingested

        :param filename: Path to a .xlsx file
        :type filename: PathLike
        :param limit: The maximum number of rows to read from each sheet. If -1 (default), reads all rows.
        :type limit: Optional[int], optional
        """

        self.filename = filename
        self.limit = limit or -1

    @cached_property
    def records(self) -> List[Record]:
        """The set of rows in all sheets in the XLSX file.

        :return: A Record representing a row in a sheet. The sheetname is stored
                in the 'rtype' field. The 'fields' field contains key-value pairs
                for each row; the keys are the names of the columns and the values
                are the cell values at that column for the given row.
        :rtype: List[Record]
        """
        records = []
        # using data_only means that the cells will contain data, rather than a formula.
        wb = load_workbook(self.filename, data_only=True)  # noqa
        for sheetname in wb.sheetnames:
            sheet = wb[sheetname]
            columns = [sheet.cell(1, c + 1).value for c in range(sheet.max_column)]
            upper_range = sheet.max_row + 1 if self.limit < 0 else self.limit
            for row in range(2, upper_range):
                fields = {
                    columns[c]: sheet.cell(row, c + 1).value
                    for c in range(sheet.max_column)
                }
                records.append(
                    Record(
                        rtype=sheetname,
                        fields=fields,
                    )
                )
        return records



import csv
import logging
import warnings
from collections import Counter
from copy import copy
from dataclasses import dataclass
from io import BytesIO, StringIO
from itertools import chain
from os import PathLike
from secrets import token_hex
from typing import TYPE_CHECKING, Dict, Generator, List, Optional, Set, Tuple, Union

import rdflib
from rdflib.term import Node

from buildingmotif import get_building_motif
from buildingmotif.dataclasses.model import Model
from buildingmotif.namespaces import bind_prefixes
from buildingmotif.template_matcher import Mapping, TemplateMatcher
from buildingmotif.utils import (
    PARAM,
    _strip_param,
    combine_graphs,
    copy_graph,
    remove_triples_with_node,
    replace_nodes,
)

if TYPE_CHECKING:
    from buildingmotif import BuildingMOTIF
    from buildingmotif.dataclasses.library import Library


@dataclass
class Template:
    """This class mirrors :py:class`database.tables.DBTemplate`."""

    _id: int
    _name: str
    body: rdflib.Graph
    optional_args: List[str]
    _bm: "BuildingMOTIF"

    @classmethod
    def load(cls, id: int) -> "Template":
        """Load template from database.

        :param id: id of template
        :type id: int
        :return: loaded template
        :rtype: Template
        """
        bm = get_building_motif()
        db_template = bm.table_connection.get_db_template(id)
        body = bm.graph_connection.get_graph(db_template.body_id)

        return cls(
            _id=db_template.id,
            _name=db_template.name,
            optional_args=db_template.optional_args,
            body=body,
            _bm=bm,
        )

    def in_memory_copy(self) -> "Template":
        """Copy this template.

        :return: copy of this template
        :rtype: Template
        """
        return Template(
            _id=-1,
            _name=self._name,
            body=copy_graph(self.body, preserve_blank_nodes=False),
            optional_args=self.optional_args[:],
            _bm=self._bm,
        )

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, new_id):
        raise AttributeError("Cannot modify db id")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name: str) -> None:
        self._bm.table_connection.update_db_template_name(self._id, new_name)
        self._name = new_name

    def get_dependencies(self) -> Tuple["Dependency", ...]:
        """Get the template's dependencies.

        :return: a tuple of dependencies
        :rtype: Tuple
        """
        return tuple(
            [
                Dependency(dep.dependee_id, dep.args)
                for dep in self._bm.table_connection.get_db_template_dependencies(
                    self._id
                )
            ]
        )

    def add_dependency(self, dependency: "Template", args: Dict[str, str]) -> None:
        """Add dependency to template.

        :param dependency: dependency to add
        :type dependency: Template
        :param args: dictionary of dependency arguments
        :type args: Dict[str, str]
        """
        self._bm.table_connection.add_template_dependency_preliminary(
            self.id, dependency.id, args
        )

    def check_dependencies(self):
        """
        Verifies that all dependencies have valid references to the parameters
        in the dependency or (recursively) its dependencies. Raises an exception if any
        issues are found.

        It is recommended to call this after *all* templates in a library have been
        loaded in to the DB, else this might falsely raise an error.
        """
        for dep in self._bm.table_connection.get_db_template_dependencies(self.id):
            self._bm.table_connection.check_template_dependency_relationship(dep)

    def remove_dependency(self, dependency: "Template") -> None:
        """Remove dependency from template.

        :param dependency: dependency to remove
        :type dependency: Template
        """
        self._bm.table_connection.delete_template_dependency(self.id, dependency.id)

    @property
    def all_parameters(self) -> Set[str]:
        """The set of all parameters used in this template *including* its
        dependencies. Includes optional parameters.

        :return: set of parameters *with* dependencies
        :rtype: Set[str]
        """
        # handle local parameters first
        params = set(self.parameters)

        # then handle dependencies
        for dep in self.get_dependencies():
            params.update(dep.template.parameters)
        return params

    @property
    def parameters(self) -> Set[str]:
        """The set of all parameters used in this template *excluding* its
        dependencies. Includes optional parameters.

        :return: set of parameters *without* dependencies
        :rtype: Set[str]
        """
        # handle local parameters first
        nodes = chain.from_iterable(self.body.triples((None, None, None)))
        params = {str(p)[len(PARAM) :] for p in nodes if str(p).startswith(PARAM)}
        return params

    @property
    def dependency_parameters(self) -> Set[str]:
        """The set of all parameters used in this template's dependencies, including
        optional parameters.

        :return: set of parameters used in dependencies
        :rtype: Set[str]
        """
        params: Set[str] = set()
        for dep in self.get_dependencies():
            params = params.union(dep.template.parameters)
        return params

    @property
    def parameter_counts(self) -> Counter:
        """An addressable histogram of the parameter name counts in this
        template and all of its transitive dependencies.

        :return: count of parameters
        :rtype: Counter
        """
        counts: Counter = Counter()
        counts.update(self.parameters)
        for dep in self.get_dependencies():
            counts.update(dep.template.parameter_counts)
        return counts

    # TODO: method to get the 'types' of the parameters

    def dependency_for_parameter(self, param: str) -> Optional["Template"]:
        """Returns the dependency that uses the given parameter if one exists.

        :param param: parameter to search for
        :type param: str
        :return: dependency that uses the given parameter
        :rtype: Optional[Template]
        """
        for dep in self.get_dependencies():
            if param in dep.args.values():
                return dep.template
        return None

    def to_inline(self, preserve_args: Optional[List[str]] = None) -> "Template":
        """Return an inline-able copy of this template.

        Suffixes all parameters with a unique identifier that will avoid
        parameter name collisions when templates are combined with one another.
        Any argument names in the `preserve_args` list will not be adjusted.

        :param preserve_args: parameters whose names will be preserved,
            defaults to None
        :type preserve_args: Optional[List[str]], optional
        :return: a template with globally unique parameters
        :rtype: Template
        """
        templ = self.in_memory_copy()
        suffix = f"{token_hex(4)}-inlined"
        # the lookup table of old to new parameter names
        to_replace = {}
        for param in templ.parameters:
            # skip if (a) we want to preserve the param or (b) it is already inlined
            if (preserve_args and param in preserve_args) or (
                param.endswith("-inlined")
            ):
                continue
            param = PARAM[param]
            to_replace[param] = rdflib.URIRef(f"{param}-{suffix}")
        replace_nodes(templ.body, to_replace)
        return templ

    @property
    def transitive_parameters(self) -> Set[str]:
        """Get all parameters used in this template and its dependencies.

        :return: set of all parameters
        :rtype: Set[str]
        """
        params = set(self.parameters)
        for dep in self.get_dependencies():
            transitive_params = dep.template.transitive_parameters
            rename_params: Dict[str, str] = {
                ours: theirs for ours, theirs in dep.args.items()
            }
            name_prefix = dep.args.get("name")
            for param in transitive_params:
                if param not in dep.args and param != "name":
                    rename_params[param] = f"{name_prefix}-{param}"
            params.update(rename_params.values())
        return params

    def inline_dependencies(self) -> "Template":
        """Copies this template with all dependencies recursively inlined.

        Parameters of dependencies will be renamed to avoid confusion.

        :return: copy of this template with all dependencies inlined
        :rtype: Template
        """
        templ = self.in_memory_copy()
        # if this template has no dependencies, then return unaltered
        if not self.get_dependencies():
            return templ

        # start with this template's parameters; this recurses into each
        # dependency to inline dependencies
        for dep in self.get_dependencies():
            # get the inlined version of the dependency
            deptempl = dep.template.inline_dependencies()

            # replace dependency parameters with the names they inherit
            # through the provided bindings
            rename_params: Dict[str, str] = {
                ours: _strip_param(theirs) for ours, theirs in dep.args.items()
            }
            # replace all parameters *not* mentioned in the args by prefixing
            # them with the 'name' parameter binding; this is guaranteed
            # to exist
            name_prefix = _strip_param(dep.args["name"])
            # for each parameter in the dependency...
            for param in deptempl.parameters:
                # if it does *not* have a mapping in the dependency, then
                # prefix the parameter with the value of the 'name' binding
                # to scope it properly
                param = _strip_param(param)
                if param not in dep.args and param != "name":
                    rename_params[param] = f"{name_prefix}-{param}"

            # replace the parameters in the dependency template
            replace_nodes(
                deptempl.body, {PARAM[k]: PARAM[v] for k, v in rename_params.items()}
            )
            # rename the optional_args in the dependency template too
            deptempl.optional_args = [
                rename_params.get(arg, arg) for arg in deptempl.optional_args
            ]

            # at this point, deptempl's parameters are all unique with respect to
            # the parent template. They are either renamed explicitly via the dependency's
            # args or implicitly via prefixing with the 'name' parameter.

            # Next, we need to determine which of deptempl's parameters are optional
            # and add these to the parent template's optional_args list.

            # get the parent template's optional args
            templ_optional_args = set(templ.optional_args)

            # represents the optional parameters of the dependency template
            deptempl_opt_args: Set[str] = set()

            # these optional parameters come from two places.
            # 1. the dependency template itself (its optional_args)
            deptempl_opt_args.update(deptempl.optional_args)
            # 1a. remove any parameters that have the same name as a parameter in the
            #     parent but are not optional in the parent
            deptempl_opt_args.difference_update(templ.parameters)
            # 2. having the same name as an optional parameter in the parent template
            #    (templ_optional_args)
            deptempl_opt_args.update(
                templ_optional_args.intersection(deptempl.parameters)
            )
            # 2a. if the 'name' of the deptempl is optional (given by the parent template),
            #   then all the arguments inside deptempl become optional
            #   (deptempl.parameters)
            if rename_params["name"] in deptempl_opt_args:
                # mark all of deptempl's parameters as optional
                deptempl_opt_args.update(deptempl.parameters)

            # convert our set of optional params to a list and assign to the parent template
            # 1. get required parameters from the original template
            # 2. calculate all optional requirements from the dependency template and the original template
            # 3. remove required parameters from the optional requirements
            # This avoids a bug where an optional dependency makes reference to a required parameter, and then
            # subsequent inlining of the dependency without optional args would remove the required parameter
            required = templ.parameters - templ_optional_args
            templ.optional_args = list(
                templ_optional_args.union(deptempl_opt_args) - required
            )

            # append the inlined template into the parent's body
            templ.body += deptempl.body

        return templ

    def evaluate(
        self,
        bindings: Dict[str, Node],
        namespaces: Optional[Dict[str, rdflib.Namespace]] = None,
        require_optional_args: bool = False,
        warn_unused: bool = True,
    ) -> Union["Template", rdflib.Graph]:
        """Evaluate the template with the provided bindings.

        If all parameters in the template have a provided binding, then a graph
        will be returned. Otherwise, a new Template will be returned that
        incorporates the provided bindings and preserves unbound parameters. If
        `require_optional_args` is True, then the template evaluation will not
        return a graph unless all optional arguments are bound. If
        `require_optional_args` is False, then the template evaluation will
        return a graph even if some optional arguments are unbound.

        :param bindings: map of parameter {name: RDF term} to substitute
        :type bindings: Dict[str, Node]
        :param namespaces: namespace bindings to add to the graph,
            defaults to None
        :type namespaces: Optional[Dict[str, rdflib.Namespace]], optional
        :param require_optional_args: whether to require all optional arguments
            to be bound, defaults to False
        :type require_optional_args: bool
        :param warn_unused: if True, print a warning if there were any parameters left
            unbound during evaluation. If 'require_optional_args' is True,
            then this will consider optional parameters when producing the warning;
            otherwise, optional paramters will be ignored; defaults to True
        :type warn_unused: bool
        :return: either a template or a graph, depending on whether all
            parameters were provided
        :rtype: Union[Template, rdflib.Graph]
        """
        # TODO: handle datatype properties
        templ = self.in_memory_copy()
        # put all of the parameter names into the PARAM namespace so they can be
        # directly subsituted in the template body
        uri_bindings: Dict[Node, Node] = {PARAM[k]: v for k, v in bindings.items()}
        # replace the param:<name> URIs in the template body with the bindings
        # provided in the call to evaluate()
        replace_nodes(templ.body, uri_bindings)
        leftover_params = (
            templ.parameters.difference(bindings.keys())
            if not require_optional_args
            else (templ.parameters.union(self.optional_args)).difference(
                bindings.keys()
            )
        )
        # true if all parameters are now bound or only optional args are unbound
        if len(templ.parameters) == 0 or (
            not require_optional_args
            and templ.parameters.issubset(set(self.optional_args))
        ):
            bind_prefixes(templ.body)
            if namespaces:
                for prefix, namespace in namespaces.items():
                    templ.body.bind(prefix, namespace)
            if not require_optional_args:
                # remove all triples that touch unbound optional_args
                unbound_optional_args = set(templ.optional_args) - set(bindings.keys())
                for arg in unbound_optional_args:
                    remove_triples_with_node(templ.body, PARAM[arg])
            return templ.body
        if len(leftover_params) > 0 and warn_unused:
            warnings.warn(
                f"Parameters \"{', '.join(leftover_params)}\" were not provided during evaluation",
                UserWarning,
            )
        return templ

    def fill(
        self, ns: rdflib.Namespace, include_optional: bool = False
    ) -> Tuple[Dict[str, Node], rdflib.Graph]:
        """Evaluates the template with autogenerated bindings within the given
        namespace.

        :param ns: namespace to contain the autogenerated entities
        :type ns: rdflib.Namespace
        :param include_optional: if True, invent URIs for optional parameters; ignore if False
        :type include_optional: bool
        :return: a tuple of the bindings used and the resulting graph
        :rtype: Tuple[Dict[str, Node], rdflib.Graph]
        """
        bindings: Dict[str, Node] = {
            param: ns[f"{param}_{token_hex(4)}"]
            for param in self.parameters
            if include_optional or param not in self.optional_args
        }
        res = self.evaluate(bindings, require_optional_args=include_optional)
        assert isinstance(res, rdflib.Graph)
        return bindings, res

    @property
    def defining_library(self) -> "Library":
        """The library defining this template.

        :return: library
        :rtype: Library
        """
        from buildingmotif.dataclasses.library import Library

        return Library.load(
            self._bm.table_connection.get_library_defining_db_template(self.id).id
        )

    def library_dependencies(self) -> List["Library"]:
        """Get library dependencies for this template.

        :return: list of libraries
        :rtype: List[Library]
        """
        from buildingmotif.dataclasses.library import Library

        libs = {self.defining_library.id}
        for dep in self.get_dependencies():
            libs.add(dep.template.defining_library.id)
        return [Library.load(id) for id in libs]

    def find_subgraphs(
        self, model: Model, *ontologies: rdflib.Graph
    ) -> Generator[Tuple[Mapping, rdflib.Graph, Optional["Template"]], None, None]:
        """Produces an iterable of subgraphs in the model that are partially or
        entirely covered by the provided template.

        :yield: iterable of subgraphs in the model
        :rtype: Generator[Tuple[Mapping, rdflib.Graph, Optional[Template]], None, None]
        """
        # TODO: can we figure out what ontology to use automatically?
        # if ontology is not specified, pull in all shapes related to this template's library
        # and all of its dependencies
        if len(ontologies) == 0:
            ontology = rdflib.Graph()
            for lib in self.library_dependencies():
                ontology += lib.get_shape_collection().graph
        else:
            ontology = combine_graphs(*ontologies)

        matcher = TemplateMatcher(model.graph, self, ontology)
        for mapping, sg in matcher.building_mapping_subgraphs_iter():
            yield mapping, sg, matcher.remaining_template(mapping)

    def generate_csv(self, path: Optional[PathLike] = None) -> Optional[StringIO]:
        """
        Generate a CSV for this template which contains a column for each template parameter.
        Once filled out, the resulting CSV file can be passed to a Template Ingress to populate a model.
        Returns a 'io.BytesIO' object which can be written to a file or sent to another program/function.

        :param path: if not None, writes the CSV to the indicated file
        :type path: PathLike, optional
        :return: String buffer containing the resulting CSV file
        :rtype: StringIO
        """
        all_parameters = copy(self.parameters)
        mandatory_parameters = all_parameters - set(self.optional_args)
        row_data = list(mandatory_parameters) + list(self.optional_args)

        if path is not None:
            # write directly to file
            with open(path, "w") as f:
                writer = csv.writer(f)
                writer.writerow(row_data)
            return None

        # write to in-memory file
        output = StringIO()
        writer = csv.writer(output)
        writer.writerow(row_data)
        return output

    def generate_spreadsheet(
        self, path: Optional[PathLike] = None
    ) -> Optional[BytesIO]:
        """
        Generate a spreadsheet for this template which contains a column for each template parameter.
        Once filled out, the resulting spreadsheet can be passed to a Template Ingress to populate a model.
        Returns a 'io.BytesIO' object which can be written to a file or sent to another program/function.

        :param path: if not None, writes the CSV to the indicated file
        :type path: PathLike, optional
        :return: Byte buffer containing the resulting spreadsheet file
        :rtype: BytesIO
        """
        try:
            from openpyxl import Workbook
            from openpyxl.utils import get_column_letter
            from openpyxl.worksheet.table import Table, TableStyleInfo
        except ImportError:
            logging.critical(
                "Install the 'xlsx-ingress' module, e.g. 'pip install buildingmotif[xlsx-ingress]'"
            )
            return None
        all_parameters = copy(self.parameters)
        mandatory_parameters = all_parameters - set(self.optional_args)

        workbook = Workbook()
        sheet = workbook.active
        if sheet is None:
            raise Exception("Could not open active sheet in Workbook")

        row_data = list(mandatory_parameters) + list(self.optional_args)
        for column_index, cell_value in enumerate(row_data, 1):
            column_letter = get_column_letter(column_index)
            sheet[f"{column_letter}1"] = cell_value  # type: ignore
            # Adjust column width based on cell content
            column_dimensions = sheet.column_dimensions[column_letter]  # type: ignore
            column_dimensions.width = max(column_dimensions.width, len(str(cell_value)))

        tab = Table(
            displayName="Table1", ref=f"A1:{get_column_letter(len(row_data))}10"
        )
        style = TableStyleInfo(name="TableStyleMedium9", showRowStripes=True)
        tab.tableStyleInfo = style
        sheet.add_table(tab)

        if path is not None:
            # write directly to file
            workbook.save(path)
            return None

        # save the file in-memory and return the resulting buffer
        f = BytesIO()
        workbook.save(f)
        return f


@dataclass
class Dependency:
    """Dependency"""

    _template_id: int
    args: Dict[str, str]

    @property
    def template_id(self):
        return self._template_id

    @property
    def template(self) -> Template:
        return Template.load(self._template_id)



import logging
import re
from typing import List

from rdflib import URIRef

from buildingmotif.label_parsing.parser import Parser
from buildingmotif.label_parsing.tokens import (
    Constant,
    Delimiter,
    Identifier,
    Null,
    Token,
    TokenOrConstructor,
    TokenResult,
    ensure_token,
)
from buildingmotif.namespaces import BRICK

logger = logging.getLogger()


class string(Parser):
    """Constructs a parser that matches a string."""

    def __init__(self, s: str, type_name: TokenOrConstructor, id=None):
        self.s = s
        self.type_name = type_name
        self.id = id

    def __call__(self, target: str) -> List[TokenResult]:
        if target.startswith(self.s):
            return [
                TokenResult(
                    self.s,
                    ensure_token(self.type_name, self.s),
                    len(self.s),
                    id=self.id,
                )
            ]
        return [
            TokenResult(
                None,
                Null(),
                0,
                f"Expected {self.s}, got {target[:len(self.s)]}",
                id=self.id,
            )
        ]


class rest(Parser):
    """Constructs a parser that matches the rest of the string."""

    def __init__(self, type_name: TokenOrConstructor, id=None):
        self.type_name = type_name
        self.id = id

    def __call__(self, target: str) -> List[TokenResult]:
        return [
            TokenResult(
                target, ensure_token(self.type_name, target), len(target), id=self.id
            )
        ]


class substring_n(Parser):
    """Constructs a parser that matches a substring of length n."""

    def __init__(self, length: int, type_name: TokenOrConstructor, id=None):
        self.length = length
        self.type_name = type_name
        self.id = id

    def __call__(self, target: str) -> List[TokenResult]:
        if len(target) >= self.length:
            value = target[: self.length]
            return [
                TokenResult(
                    value, ensure_token(self.type_name, value), self.length, id=self.id
                )
            ]
        return [
            TokenResult(
                None,
                Null(),
                0,
                f"Expected {self.length} characters, got {target[:self.length]}",
                id=self.id,
            )
        ]


class regex(Parser):
    """Constructs a parser that matches a regular expression."""

    def __init__(self, r: str, type_name: TokenOrConstructor, id=None):
        self.r = r
        self.type_name = type_name
        self.id = id

    def __call__(self, target: str) -> List[TokenResult]:
        match = re.match(self.r, target)
        if match:
            value = match.group()
            return [
                TokenResult(
                    value, ensure_token(self.type_name, value), len(value), id=self.id
                )
            ]
        return [
            TokenResult(
                None,
                Null(),
                0,
                f"Expected {self.r}, got {target[:len(self.r)]}",
                id=self.id,
            )
        ]


class choice(Parser):
    """Constructs a choice combinator of parsers."""

    def __init__(self, *parsers: Parser, id=None):
        self.parsers = parsers
        self.id = id

    def __call__(self, target: str) -> List[TokenResult]:
        errors = []
        for p in self.parsers:
            result = p(target)
            if result and not any(r.error for r in result):
                return result
            if result:
                errors.append(result[0].error)
        return [TokenResult(None, Null(), 0, " | ".join(errors), id=None)]  # type: ignore


class constant(Parser):
    """Matches a constant token."""

    def __init__(self, type_name: Token, id=None):
        self.id = id
        self.type_name = type_name

    def __call__(self, target: str) -> List[TokenResult]:
        return [TokenResult(None, self.type_name, 0, id=self.id)]


class abbreviations(Parser):
    """Constructs a choice combinator of string matching based on a dictionary."""

    def __init__(self, patterns: dict, id=None):
        parsers = [string(s, Constant(URIRef(t))) for s, t in patterns.items()]
        self.choice = choice(*parsers)
        self.id = id

    def __call__(self, target: str):
        return self.choice(target)


class sequence(Parser):
    """Applies parsers in sequence. All parsers must match consecutively."""

    def __init__(self, *parsers: Parser, id=None):
        self.parsers = parsers
        self.id = id

    def __call__(self, target: str) -> List[TokenResult]:
        results = []
        total_length = 0
        for p in self.parsers:
            result = p(target)
            if not result:
                raise Exception("Expected result")
            results.extend(result)
            # if there are any errors, return the results
            if any(r.error for r in result):
                return results
            # TODO: how to handle error?
            consumed_length = sum([r.length for r in result])
            target = target[consumed_length:]
            total_length += sum([r.length for r in result])
        return results


class many(Parser):
    """Applies the given sequence parser repeatedly until it stops matching."""

    def __init__(self, seq_parser: Parser, id=None):
        self.seq_parser = seq_parser
        self.id = id

    def __call__(self, target):
        results = []
        while True:
            part = self.seq_parser(target)
            if not part or part[0].value is None:
                break
            results.extend(part)
            # add up the length of all the tokens
            total_length = sum([r.length for r in part])
            target = target[total_length:]
        return results


class maybe(Parser):
    """Applies the given parser, but does not fail if it does not match."""

    def __init__(self, parser: Parser, id=None):
        self.parser = parser
        self.id = id

    def __call__(self, target):
        result = self.parser(target)
        # if the result is not empty and there are no errors, return the result, otherwise return a null token
        if result and not any(r.error for r in result):
            return result
        return [TokenResult(None, Null(), 0, id=self.id)]


class until(Parser):
    """
    Constructs a parser that matches everything until the given parser matches.
    STarts with a string length of 1 and increments it until the parser matches.
    """

    def __init__(self, parser: Parser, type_name: TokenOrConstructor, id=None):
        self.type_name = type_name
        self.parser = parser
        self.id = id

    def __call__(self, target):
        length = 1
        while length <= len(target):
            result = self.parser(target[length:])
            if result and not any(r.error for r in result):
                return [
                    TokenResult(
                        target[:length],
                        ensure_token(self.type_name, target[:length]),
                        length,
                        id=self.id,
                    )
                ]
            length += 1
        return [
            TokenResult(
                None,
                Null(),
                0,
                f"Expected {self.type_name}, got {target[:length]}",
                id=self.id,
            )
        ]


class extend_if_match(Parser):
    """Adds the type to the token result."""

    def __init__(self, parser: Parser, type_name: Token, id=None):
        self.parser = parser
        self.type_name = type_name
        self.id = id

    def __call__(self, target):
        result = self.parser(target)
        if result and not any(r.error for r in result):
            result.extend([TokenResult(None, self.type_name, 0, id=self.id)])
            return result
        return result


def as_identifier(parser):
    """
    If the parser matches, add a new Identifier token after
    every Constant token in the result. The Identifier token
    has the same string value as the Constant token.
    """

    def as_identifier_parser(target):
        result = parser(target)
        if result and not any(r.error for r in result):
            new_result = []
            for r in result:
                new_result.append(r)
                if isinstance(r.token, Constant):
                    # length of the new token must be given as 0 so that the substring
                    # is not double counted
                    new_result.append(TokenResult(r.value, Identifier(r.value), 0))
            return new_result
        return result

    return as_identifier_parser


COMMON_EQUIP_ABBREVIATIONS_BRICK = {
    "AHU": BRICK.Air_Handling_Unit,
    "FCU": BRICK.Fan_Coil_Unit,
    "VAV": BRICK.Variable_Air_Volume_Box,
    "CRAC": BRICK.Computer_Room_Air_Conditioner,
    "HX": BRICK.Heat_Exchanger,
    "PMP": BRICK.Pump,
    "RVAV": BRICK.Variable_Air_Volume_Box_With_Reheat,
    "HP": BRICK.Heat_Pump,
    "RTU": BRICK.Rooftop_Unit,
    "DMP": BRICK.Damper,
    "STS": BRICK.Status,
    "VLV": BRICK.Valve,
    "CHVLV": BRICK.Chilled_Water_Valve,
    "HWVLV": BRICK.Hot_Water_Valve,
    "VFD": BRICK.Variable_Frequency_Drive,
    "CT": BRICK.Cooling_Tower,
    "MAU": BRICK.Makeup_Air_Unit,
    "R": BRICK.Room,
    "A": BRICK.Air_Handling_Unit,
}

COMMON_POINT_ABBREVIATIONS = {
    "ART": BRICK.Room_Temperature_Sensor,
    "TSP": BRICK.Air_Temperature_Setpoint,
    "HSP": BRICK.Air_Temperature_Heating_Setpoint,
    "CSP": BRICK.Air_Temperature_Cooling_Setpoint,
    "SP": BRICK.Setpoint,
    "CHWST": BRICK.Leaving_Chilled_Water_Temperature_Sensor,
    "CHWRT": BRICK.Entering_Chilled_Water_Temperature_Sensor,
    "HWST": BRICK.Leaving_Hot_Water_Temperature_Sensor,
    "HWRT": BRICK.Entering_Hot_Water_Temperature_Sensor,
    "CO": BRICK.CO_Sensor,
    "CO2": BRICK.CO2_Sensor,
    "T": BRICK.Temperature_Sensor,
    "FS": BRICK.Flow_Sensor,
    "PS": BRICK.Pressure_Sensor,
    "DPS": BRICK.Differential_Pressure_Sensor,
}

COMMON_ABBREVIATIONS = abbreviations(
    {**COMMON_EQUIP_ABBREVIATIONS_BRICK, **COMMON_POINT_ABBREVIATIONS}
)


# common parser combinators
equip_abbreviations = abbreviations(COMMON_EQUIP_ABBREVIATIONS_BRICK)
point_abbreviations = abbreviations(COMMON_POINT_ABBREVIATIONS)
delimiters = regex(r"[._:/\- ]", Delimiter)
identifier = regex(r"[a-zA-Z0-9]+", Identifier)
named_equip = sequence(equip_abbreviations, maybe(delimiters), identifier)
named_point = sequence(point_abbreviations, maybe(delimiters), identifier)



import logging
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from rdflib.graph import Graph, Store, URIRef, plugin
from rdflib.namespace import NamespaceManager

if TYPE_CHECKING:
    from buildingmotif.building_motif.building_motif import BuildingMotifEngine

PROJECT_DIR = Path(__file__).resolve().parent


class GraphConnection:
    """Manages graph connection."""

    def __init__(
        self,
        engine: "BuildingMotifEngine",
        db_identifier: Optional[str] = "buildingmotif_store",
    ) -> None:
        """Constructor for the database and datastore.

        :param engine: database engine
        :type engine: Engine
        :param db_identifier: defaults to "buildingmotif_store"
        :type db_identifier: Optional[str], optional
        """
        self.logger = logging.getLogger(__name__)

        self.store = plugin.get("SQLAlchemy", Store)(
            identifier=db_identifier, engine=engine
        )

        # avoids the warnings raised by the issue in https://github.com/RDFLib/rdflib/issues/1880
        # Eventually will require rdflib-sqlalchemy to support the 'override' keyword
        def fixed_bind(self, prefix: str, namespace: URIRef, override: bool):
            self.store.bind(prefix, namespace)

        setattr(NamespaceManager, "_store_bind", fixed_bind)

        self.logger.debug("Creating tables for graph storage")
        self.store.create_all()

    def create_graph(self, identifier: str, graph: Graph) -> Graph:
        """Create a graph in the database.

        :param identifier: identifier of graph
        :type identifier: str
        :param graph: graph to add, defaults to None
        :type graph: Graph
        :return: graph added
        :rtype: Graph
        """
        self.logger.debug(
            f"Creating graph: '{identifier}' in database with: {len(graph)} triples"
        )
        g = Graph(self.store, identifier=identifier)
        new_triples = [(s, o, p, g) for (s, o, p) in graph]
        g.addN(new_triples)

        return g

    def get_all_graph_identifiers(self) -> List[str]:
        """Get all graph identifiers.

        :return: all graph identifiers
        :rtype: List[str]
        """
        graph_identifiers = [str(c) for c in self.store.contexts()]
        return graph_identifiers

    def get_graph(self, identifier: str) -> Graph:
        """Get graph by identifier. Graph has triples, no context.

        :param identifier: graph identifier
        :type identifier: str
        :return: graph without context
        :rtype: Graph
        """
        result = Graph(self.store, identifier=identifier)
        # we used to bind prefixes here but this is unnecessary because
        # the graph has prefixes bound when it is saved

        return result

    def delete_graph(self, identifier: str) -> None:
        """Delete graph.

        :param identifier: graph identifier
        :type identifier: str
        """
        self.logger.debug(f"Deleting graph: '{identifier}'")
        g = Graph(self.store, identifier=identifier)
        self.store.remove((None, None, None), g)



import logging
import uuid
from functools import lru_cache
from typing import Dict, List, Tuple

from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound

from buildingmotif.database.errors import (
    LibraryNotFound,
    ModelNotFound,
    ShapeCollectionNotFound,
    TemplateNotFound,
)
from buildingmotif.database.tables import (
    DBLibrary,
    DBModel,
    DBShapeCollection,
    DBTemplate,
    DepsAssociation,
)


class TableConnection:
    """Controls interaction with the database."""

    def __init__(self, engine: Engine, bm) -> None:
        """Class constructor.

        :param engine: database engine
        :type engine: Engine
        :param bm: contains the session to use
        :type bm: BuildingMOTIF
        """
        self.logger = logging.getLogger(__name__)
        self.bm = bm

    # model functions

    def create_db_model(self, name: str, description: str = "") -> DBModel:
        """Create a database model.

        :param name: name of DBModel
        :type name: str
        :param description: description of the model
        :type description: str
        :return: DBModel
        :rtype: DBModel
        """
        manifest_id = str(uuid.uuid4())
        self.logger.debug(f"Creating shape collection in model: '{name}'")
        manifest = DBShapeCollection(graph_id=manifest_id)
        self.bm.session.add(manifest)

        graph_id = str(uuid.uuid4())
        self.logger.debug(f"Creating model: '{name}', with graph: '{graph_id}'")
        db_model = DBModel(
            name=name,
            graph_id=graph_id,
            description=description,
            manifest=manifest,
        )
        self.bm.session.add(db_model)

        self.bm.session.flush()

        return db_model

    def get_all_db_models(self) -> List[DBModel]:
        """Get all database models.

        :return: all DBModels
        :rtype: List[DBModel]
        """
        return self.bm.session.query(DBModel).all()

    def get_db_model(self, id: int) -> DBModel:
        """Get database model by id.

        :param id: id of DBModel
        :type id: int
        :return: DBModel
        :rtype: DBModel
        """
        try:
            db_model = self.bm.session.query(DBModel).filter(DBModel.id == id).one()
        except NoResultFound:
            raise ModelNotFound(idnum=id)
        return db_model

    def get_db_model_by_name(self, name: str) -> DBModel:
        """Get database model by name.

        :param name: name of DBModel
        :type name: str
        :return: DBModel
        :rtype: DBModel
        """
        db_model = self.bm.session.query(DBModel).filter(DBModel.name == name).one()
        return db_model

    def update_db_model_name(self, id: int, name: str) -> None:
        """Update database model name.

        :param id: id of DBModel
        :type id: int
        :param name: new name
        :type name: str
        """
        db_model = self.get_db_model(id)
        self.logger.debug(f"Updating model name from: '{db_model.name}' to: '{name}'")
        db_model.name = name

    def update_db_model_description(self, id: int, description: str) -> None:
        """Update database model description.

        :param id: id of DBModel
        :type id: int
        :param description: new description
        :type description: str
        """
        db_model = self.get_db_model(id)
        self.logger.debug(
            f"Updating model description from: '{db_model.description}' to: '{description}'"
        )
        db_model.description = description

    def delete_db_model(self, id: int) -> None:
        """Delete database model.

        :param id: id of deleted DBModel
        :type id: int
        """

        db_model = self.get_db_model(id)
        self.logger.debug(f"Deleting model: '{db_model.name}'")
        self.bm.session.delete(db_model)

    # shape collection functions
    def create_db_shape_collection(self) -> DBShapeCollection:
        """Create a database shape collection.

        :return: DBShapeCollection
        :rtype: DBShapeCollection
        """
        db_shape_collection = DBShapeCollection(graph_id=str(uuid.uuid4()))

        self.bm.session.add(db_shape_collection)
        self.bm.session.flush()

        return db_shape_collection

    def get_all_db_shape_collections(self) -> List[DBShapeCollection]:
        """Get all database shape collections.

        :return: all DBShapeCollections
        :rtype: List[DBShapeCollection]
        """
        return self.bm.session.query(DBShapeCollection).all()

    def get_db_shape_collection(self, id: int) -> DBShapeCollection:
        """Get database shape collection by id.

        :param id: id of DBShapeCollection
        :type id: int
        :return: DBShapeCollection
        :rtype: DBShapeCollection
        """
        try:
            return (
                self.bm.session.query(DBShapeCollection)
                .filter(DBShapeCollection.id == id)
                .one()
            )
        except NoResultFound:
            raise ShapeCollectionNotFound(idnum=id)

    def delete_db_shape_collection(self, id: int) -> None:
        """Delete database shape collection.

        :param id: id of deleted DBShapeCollection
        :type id: int
        """
        db_shape_collection = self.get_db_shape_collection(id)
        self.bm.session.delete(db_shape_collection)

    # library functions

    def create_db_library(self, name: str) -> DBLibrary:
        """Create database library.

        :param name: name of DBLibrary
        :type name: str
        :return: DBLibrary
        :rtype: DBLibrary
        """
        self.logger.debug(f"Creating shape collection in library: '{name}'")
        shape_collection = DBShapeCollection(graph_id=str(uuid.uuid4()))
        self.bm.session.add(shape_collection)

        self.logger.debug(f"Creating database library: '{name}'")
        library = DBLibrary(name=name, shape_collection=shape_collection)
        self.bm.session.add(library)

        self.bm.session.flush()

        return library

    def get_all_db_libraries(self) -> List[DBLibrary]:
        """Get all database libraries.

        :return: all DBLibrary
        :rtype: List[DBLibrary]
        """
        db_libraries = self.bm.session.query(DBLibrary).all()
        return db_libraries

    def get_db_library(self, id: int) -> DBLibrary:
        """Get database library by id.

        :param id: id of DBLibrary
        :type id: int
        :return: DBLibrary
        :rtype: DBLibrary
        """
        try:
            db_library = (
                self.bm.session.query(DBLibrary).filter(DBLibrary.id == id).one()
            )
        except NoResultFound:
            raise LibraryNotFound(idnum=id)
        return db_library

    def get_db_library_by_name(self, name: str) -> DBLibrary:
        """Get database library by name.

        :param name: name of DBLibrary
        :type name: str
        :return: DBLibrary
        :rtype: DBLibrary
        """
        try:
            return self.bm.session.query(DBLibrary).filter(DBLibrary.name == name).one()
        except NoResultFound:
            raise LibraryNotFound(name=name)

    def update_db_library_name(self, id: int, name: str) -> None:
        """Update database library name.

        :param id: id of DBLibrary
        :type id: int
        :param name: new name
        :type name: str
        """
        db_library = self.get_db_library(id)
        self.logger.debug(
            f"Updating database library name: '{db_library.name}' -> '{name}'"
        )
        db_library.name = name

    def delete_db_library(self, id: int) -> None:
        """Delete database library.

        :param id: id of deleted DBLibrary
        :type id: int
        """

        db_library = self.get_db_library(id)

        self.logger.debug(f"Deleting database library: '{db_library.name}'")
        self.bm.session.delete(db_library)

    # template functions

    def create_db_template(self, name: str, library_id: int) -> DBTemplate:
        """Create database template.

        :param name: name of DBTemplate
        :type name: str
        :param library_id: id of the template's library
        :type library_id: int
        :return: DBTemplate
        :rtype: DBTemplate
        """
        self.logger.debug(f"Creating database template: '{name}'")
        library = self.get_db_library(library_id)
        template = DBTemplate(
            name=name,
            body_id=str(uuid.uuid4()),
            optional_args=[],
            library=library,
        )

        self.bm.session.add(template)
        self.bm.session.flush()

        return template

    def get_all_db_templates(self) -> List[DBTemplate]:
        """Get all database templates.

        :return: all DBTemplate
        :rtype: List[DBTemplate]
        """
        db_templates = self.bm.session.query(DBTemplate).all()
        return db_templates

    def get_db_template(self, id: int) -> DBTemplate:
        """Get database template by id.

        :param id: id of DBTemplate
        :type id: int
        :return: DBTemplate
        :rtype: DBTemplate
        """
        try:
            db_template = (
                self.bm.session.query(DBTemplate).filter(DBTemplate.id == id).one()
            )
        except NoResultFound:
            raise TemplateNotFound(idnum=id)
        return db_template

    def get_db_template_by_name(self, name: str) -> DBTemplate:
        """Get database template by name.

        :param name: name of DBTemplate
        :type name: str
        :return: DBTemplate
        :rtype: DBTemplate
        """
        try:
            db_template = (
                self.bm.session.query(DBTemplate).filter(DBTemplate.name == name).one()
            )
        except NoResultFound:
            raise TemplateNotFound(name=name)
        return db_template

    def get_library_defining_db_template(self, id: int) -> DBLibrary:
        """Returns the library defining the given template.

        :param id: id of template
        :type id: int
        :return: DBLibrary
        :rtype: DBLibrary
        """
        return self.get_db_template(id).library

    def get_db_template_dependencies(self, id: int) -> Tuple[DepsAssociation, ...]:
        """Get a template's dependencies and its arguments.

        If you don't need the arguments, consider using
        :py:method:`dataclasses.template.get_dependencies`.

        :param id: template id
        :type id: int
        :return: tuple of tuple, where each tuple has the dependant_id
            and it's args
        :rtype: tuple[tuple[int, list[str]]]
        """
        db_template_dependencies = tuple(
            self.bm.session.query(DepsAssociation)
            .filter(DepsAssociation.dependant_id == id)
            .all()
        )
        return db_template_dependencies

    def update_db_template_name(self, id: int, name: str) -> None:
        """Update database template name.

        :param id: id of DBTemplate
        :type id: int
        :param name: new name
        :type name: str
        """
        db_template = self.get_db_template(id)
        self.logger.debug(
            f"Updating database template name: '{db_template.name}' -> '{name}'"
        )
        db_template.name = name

    def update_db_template_optional_args(
        self, id: int, optional_args: List[str]
    ) -> None:
        """Update database template optional arguments.

        :param id: id of DBTemplate
        :type id: int
        :param optional_args: new list of optional_args
        :type name: List[str]
        """
        db_template = (
            self.bm.session.query(DBTemplate).filter(DBTemplate.id == id).one()
        )
        db_template.optional_args = optional_args

    def add_template_dependency_preliminary(
        self, template_id: int, dependency_id: int, args: Dict[str, str]
    ):
        """Creates a *preliminary* dependency between two templates. This dependency
        is preliminary because the bindings between the dependent/dependency templates
        are *not validated*. This serves to populate the directed acyclic graph of dependencies between templates
        before the parameter bindings are checked. This ensures that all of the parameters for
        a template *and those in its dependencies* can all be identified and used as part of
        the parameter binding check. The upshot of this process is dependencies between two
        templates can refer to parameters in a template *or its dependencies*. This is necessary
        to support templates that contain many nested components that refer to each other (such
        as the s223:mapsTo property).

        **Important:** Be sure to run "check_all_template_dependencies" to ensure all of your templates.
        will work as you expect!

        :param template_id: dependant template id
        :type template_id: int
        :param dependency_id: dependency template id
        :type dependency_id: int
        :param args: mapping of dependency params to dependant params
        :type args: Dict[str, str]
        :raises ValueError: if all dependee required_params not in args
        :raises ValueError: if dependant and dependency template don't share a
        """
        self.logger.debug(
            f"Creating depencency from templates with ids: '{template_id}' and: '{dependency_id}'"
        )
        templ = self.get_db_template(template_id)
        if "name" not in args.keys():
            raise ValueError(
                f"The name parameter is required for the dependency '{templ.name}'."
            )
        # In the past we had a check here to make sure the two templates were in the same library.
        # This has been removed because it wasn't actually necessary, but we may add it back in
        # in the future.
        relationship = DepsAssociation(
            dependant_id=template_id,
            dependee_id=dependency_id,
            args=args,
        )

        self.bm.session.add(relationship)
        self.bm.session.flush()

    def check_all_template_dependencies(self):
        """
        Verifies that all template dependencies have valid references to the parameters
        in the dependency or (recursively) its dependencies. Raises an exception if any
        issues are found. Checks all templates in the database.
        """
        # TODO: maybe optimize this to only check recently added templates
        # (i.e. maybe those added since last commit?)
        for db_templ in self.get_all_db_templates():
            for dep in self.get_db_template_dependencies(db_templ.id):
                self.check_template_dependency_relationship(dep)

    @lru_cache(maxsize=128)
    def check_template_dependency_relationship(self, dep: DepsAssociation):
        """Verify that the dependency between two templates is well-formed. This involves
        a series of checks:
        - existence of the dependent and dependency templates is performed during the
        :py:method:`database.template_connection.`add_template_dependency_preliminary` method
        - the args keys appear in the dependency, or recursively in a template that is a dependency
        of the named dependency
        - the args values appear in the dependent template
        - there is a 'name' parameter in the dependent template

        :param dep: the dependency object to check
        :type dep: DepsAssociation
        :raises ValueError: if all dependee required_params not in args
        :raises ValueError: if dependant and dependency template don't share a
            library
        """
        template_id = dep.dependant_id
        dependency_id = dep.dependee_id
        args = dep.args
        self.logger.debug(
            f"Creating depencency from templates with ids: '{template_id}' and: '{dependency_id}'"
        )
        from buildingmotif.dataclasses import Template

        templ = Template.load(template_id)
        params = templ.transitive_parameters
        dep_templ = Template.load(dependency_id)
        dep_params = dep_templ.transitive_parameters

        # check parameters are valid
        if not set(args.values()).issubset(params):
            raise ValueError(
                f"In {templ.name} the values of the bindings to {dep_templ.name} must correspond to the "
                "parameters in the dependant template."
                f"Dependency {dep_templ.name} refers to params {set(args.values()).difference(params)} "
                f"that do not appear in template {templ.name} ({params})."
            )
        # do the same for the dependency
        binding_params = set(args.keys())
        if not binding_params.issubset(dep_params):
            diff = binding_params.difference(dep_params)
            raise ValueError(
                f"In {templ.name} the keys of the bindings to {dep_templ.name} must correspond to the "
                "parameters in the template dependency. "
                f"The dependency binding uses params {diff} which don't appear in the dependency template. "
                f"Available dependency parameters are {dep_params}"
            )

        # TODO: do we need this kind of check?
        if len(params) > 0 and args["name"] not in params:
            raise ValueError(
                "The name parameter of the dependency must be bound to a param in this template."
                f"'name' was bound to {args['name']} but available params are {params}"
            )

    def delete_template_dependency(self, template_id: int, dependency_id: int):
        """Delete dependency between two templates.

        :param template_id: dependant template id
        :type template_id: int
        :param dependency_id: dependency template id
        :type dependency_id: int
        """
        self.logger.debug(
            f"Deleting depencency from templates with ids: '{template_id}' and: '{dependency_id}'"  # noqa
        )

        relationship = (
            self.bm.session.query(DepsAssociation)
            .filter(
                DepsAssociation.dependant_id == template_id,
                DepsAssociation.dependee_id == dependency_id,
            )
            .one()
        )
        self.bm.session.delete(relationship)

    def update_db_template_library(self, id: int, library_id: int) -> None:
        """Update database template library.

        :param id: id of DBTemplate
        :type id: int
        :param library_id: id of the new library
        :type library_id: int
        """
        db_template = self.get_db_template(id)
        self.logger.debug(
            f"Updating database template library: '{db_template.library_id}' -> '{library_id}'"  # noqa
        )
        db_template.library_id = library_id

    def delete_db_template(self, id: int) -> None:
        """Delete database template.

        :param id: id of deleted DBTemplate
        :type id: int
        """
        db_template = self.get_db_template(id)
        self.logger.debug(f"Deleting template: '{db_template.name}'")

        self.bm.session.delete(db_template)



import os
from typing import Optional

from flask import Flask, current_app
from flask_api import status
from sqlalchemy.exc import SQLAlchemyError

from buildingmotif.api.views.library import blueprint as library_blueprint
from buildingmotif.api.views.model import blueprint as model_blueprint
from buildingmotif.api.views.parser import blueprint as parsers_blueprint
from buildingmotif.api.views.template import blueprint as template_blueprint
from buildingmotif.building_motif.building_motif import BuildingMOTIF


def _after_request(response):
    """Commit or rollback the session.

    :param response: response
    :type response: Flask.response
    :return: response
    :rtype: Flask.response
    """
    try:
        current_app.building_motif.session.commit()
    except SQLAlchemyError:
        current_app.building_motif.session.rollback()

    current_app.building_motif.Session.remove()
    response.headers["Access-Control-Allow-Origin"] = "*"
    response.headers["Access-Control-Allow-Methods"] = "*"
    response.headers["Access-Control-Allow-Headers"] = "*"

    return response


def _after_error(error):
    """Returns request with a 500 and the error message.

    :param error: python error
    :type error: Error
    :return: flask error response
    :rtype: Flask.response
    """
    return str(error), status.HTTP_500_INTERNAL_SERVER_ERROR


def create_app(DB_URI, shacl_engine: Optional[str] = "pyshacl"):
    """Creates a Flask API.

    :param db_uri: database URI
    :type db_uri: str
    :param shacl_engine: the name of the engine to use for validation: "pyshacl" or "topquadrant". Using topquadrant
        requires Java to be installed on this machine, and the "topquadrant" feature on BuildingMOTIF,
        defaults to "pyshacl"
    :type shacl_engine: str, optional
    :return: flask app
    :rtype: Flask.app
    """
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_mapping(
        DB_URI=DB_URI,
    )
    app.building_motif = BuildingMOTIF(app.config["DB_URI"], shacl_engine=shacl_engine)

    app.after_request(_after_request)
    app.register_error_handler(Exception, _after_error)

    app.register_blueprint(library_blueprint, url_prefix="/libraries")
    app.register_blueprint(template_blueprint, url_prefix="/templates")
    app.register_blueprint(model_blueprint, url_prefix="/models")
    app.register_blueprint(parsers_blueprint, url_prefix="/parsers")

    return app


if __name__ == "__main__":
    """Run API."""
    db_uri = os.getenv("DB_URI")
    if db_uri is None:
        raise ValueError("Environment variable DB_URI not set.")

    app = create_app(db_uri)
    app.run(debug=True, host="0.0.0.0", threaded=False)



from buildingmotif.dataclasses.library import Library  # noqa
from buildingmotif.dataclasses.model import Model  # noqa
from buildingmotif.dataclasses.shape_collection import ShapeCollection  # noqa
from buildingmotif.dataclasses.template import Template  # noqa
from buildingmotif.dataclasses.validation import ValidationContext  # noqa



import secrets
from typing import Dict, List, Optional, Union

from rdflib import BNode, Graph, Literal, Namespace, URIRef
from rdflib.store import Store
from rdflib.term import Node

from buildingmotif.dataclasses import Library, Template
from buildingmotif.namespaces import RDF, RDFS


class TemplateBuilderContext(Graph):
    """
    A context for building templates. This class allows the user to
    add templates to the context and then access them by name. The
    context also allows the user to compile all of the templates in
    the context into a single graph.
    """

    def __init__(self, ns: Namespace, store: Optional[Union[Store, str]] = None):
        """
        Creates a new TemplateBuilderContext. The context will create
        entities in the given namespace.

        :param ns:  The namespace to use for the context
        :param store: An optional backing store for the graph; ok to leave blank unless
                    you are experiencing performance issues using TemplateBuilderContext
        """
        self.templates: Dict[str, Template] = {}
        self.wrappers: List[TemplateWrapper] = []
        self.ns: Namespace = ns
        super(TemplateBuilderContext, self).__init__(
            store=store or "default", identifier=None
        )

    def add_template(self, template: Template):
        """
        Adds a template to the context with all of its dependencies
        inlined. Allows the user of the context to access the template
        by name.

        :param template: The template to add to the context
        """
        self.templates[template.name] = template.inline_dependencies()

    def add_templates_from_library(self, library: Library):
        """
        Adds all of the templates from a library to the context

        :param library: The library to add to the context
        """
        for template in library.get_templates():
            self.add_template(template)

    def __getitem__(self, template_name):
        if template_name in self.templates:
            w = TemplateWrapper(self.templates[template_name], self.ns)
            self.wrappers.append(w)
            return w
        else:
            raise KeyError(f"Invalid template name: {template_name}")

    def compile(self) -> Graph:
        """
        Compiles all of the template wrappers and concatenates them into a single Graph

        :return: A graph containing all of the compiled templates
        """
        graph = Graph()
        graph += self
        for wrapper in self.wrappers:
            graph += wrapper.compile()
        # add a label to every instance if it doesn't have one. Make
        # the label the same as the value part of the URI
        for s, o in graph.subject_objects(predicate=RDF.type):
            if (s, RDFS.label, None) not in graph:
                # get the 'value' part of the o URI using qname
                _, _, value = graph.namespace_manager.compute_qname(str(o))
                graph.add((s, RDFS.label, Literal(value)))
        return graph


class TemplateWrapper:
    def __init__(self, template: Template, ns: Namespace):
        """
        Creates a new TemplateWrapper. The wrapper is used to bind
        parameters to a template and then compile the template into
        a graph.

        :param template: The template to wrap
        :param ns: The namespace to use for the wrapper; all bindings will be added to this namespace
        """
        self.template = template
        self.bindings: Dict[str, Node] = {}
        self.ns = ns

    def __call__(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v
        return self

    def __getitem__(self, param):
        if param in self.bindings:
            return self.bindings[param]
        elif param not in self.template.all_parameters:
            raise KeyError(f"Invalid parameter: {param}")
        # if the param is not bound, then invent a name
        # by prepending the parameter name to a random string
        self.bindings[param] = self.ns[param + "_" + secrets.token_hex(4)]
        return self.bindings[param]

    def __setitem__(self, param, value):
        if param not in self.template.all_parameters:
            raise KeyError(f"Invalid parameter: {param}")
        # if value is not a URIRef, Literal or BNode, then put it in the namespace
        if not isinstance(value, (URIRef, Literal, BNode)):
            value = self.ns[value]
        # check datatype of value is URIRef, Literal or BNode
        if not isinstance(value, (URIRef, Literal, BNode)):
            raise TypeError(f"Invalid type for value: {type(value)}")
        self.bindings[param] = value

    @property
    def parameters(self):
        return self.template.parameters

    def compile(self) -> Graph:
        """
        Compiles the template into a graph. If there are still parameters
        to be bound, then the template will be returned. Otherwise, the
        template will be filled and the resulting graph will be returned.

        :return: A graph containing the compiled template
        :rtype: Graph
        """
        tmp = self.template.evaluate(self.bindings)
        # if this is true, there are still parameters to be bound
        if isinstance(tmp, Template):
            bindings, graph = tmp.fill(self.ns, include_optional=False)
            self.bindings.update(bindings)
            return graph
        else:
            return tmp



from typing import Any

import jsonschema

LIBRARIES_YAML_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "oneOf": [
            {
                "type": "object",
                "properties": {"directory": {"type": "string"}},
                "required": ["directory"],
            },
            {
                "type": "object",
                "properties": {"ontology": {"type": "string"}},
                "required": ["ontology"],
            },
            {
                "type": "object",
                "properties": {
                    "git": {
                        "type": "object",
                        "properties": {
                            "repo": {"type": "string"},
                            "branch": {"type": "string"},
                            "path": {"type": "string"},
                        },
                        "required": ["repo", "branch", "path"],
                    }
                },
                "required": ["git"],
            },
        ],
    },
}


def validate_libraries_yaml(doc: Any):
    """
    Validates a given document against the library.yml schema. Raises
    a jsonschema.exceptions.ValidationError if errors are found

    :param doc: a value retrieved from deserializing libraries.yml file
    """
    jsonschema.validate(schema=LIBRARIES_YAML_SCHEMA, instance=doc)



from rdflib import Graph, Literal, Namespace

from buildingmotif import BuildingMOTIF
from buildingmotif.dataclasses import Library
from buildingmotif.ingresses.bacnet import BACnetNetwork
from buildingmotif.ingresses.base import GraphIngressHandler


def _clean_uri(n) -> str:
    if isinstance(n, str):
        return n.replace(" ", "_")
    return str(n)


class BACnetToBrickIngress(GraphIngressHandler):
    """Turns a BACnetNetwork RecordIngressHandler into a Brick model"""

    BNS = Namespace("urn:brick_bacnet_scan/")

    def __init__(self, bm: BuildingMOTIF, upstream: BACnetNetwork):
        """Create a new ignress handler for turning a BACnet network scrape
        into a Brick model

        :param bm: BuildingMOTIF instance
        :type bm: BuildingMOTIF
        :param upstream: the BACnetNetwork ingress handler to ingest from
        :type upstream: BACnetNetwork
        """
        super().__init__(bm)
        self.upstream = upstream
        self.bacnet_lib = Library.load(directory="bacnet")
        self.device_template = self.bacnet_lib.get_template_by_name("brick-device")
        self.object_template = self.bacnet_lib.get_template_by_name("brick-point")

    def graph(self, ns: Namespace) -> Graph:
        """Generates a Brick graph from the BACnet network with all entities
        placed in the given namespace.

        :param ns: Namespace for all inferred entities
        :type ns: Namespace
        :return: RDF graph containing a Brick model of the BACnet network
        :rtype: Graph
        """
        g = Graph()
        # ensure 'ns' is a Namespace or URI forming won't work
        if not isinstance(ns, Namespace):
            ns = Namespace(ns)
        records = self.upstream.records
        assert records is not None
        for record in records:
            if record.rtype == "Device":
                dev = record.fields
                device_id = dev["device_id"]
                name = _clean_uri(device_id) or _clean_uri(dev["address"])
                dev_graph = self.device_template.evaluate(
                    {
                        "name": ns[name],
                        "instance-number": Literal(device_id),
                        "address": Literal(dev["address"]),
                    }
                )
                assert isinstance(dev_graph, Graph)
                g += dev_graph
            elif record.rtype == "Object":
                point = record.fields
                device_id = point["device_id"]
                obj_graph = self.object_template.evaluate(
                    {
                        "name": ns[f"{_clean_uri(point['name'])}-{point['address']}"],
                        "identifier": Literal(f"{point['type']},{point['address']}"),
                        "obj-name": Literal(point["name"]),
                        "device": ns[_clean_uri(device_id)],
                    }
                )
                assert isinstance(obj_graph, Graph)
                g += obj_graph

        return g



import json

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB


class JSONType(sa.types.TypeDecorator):
    """Custom JSON type that uses JSONB on Postgres and JSON on other dialects.
    This allows us to use our custom JSON serialization below *and* have the
    database enforce uniqueness on JSON-encoded dictionaries
    """

    impl = sa.JSON
    hashable = False
    cache_ok = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            # Use the native JSON type.
            return dialect.type_descriptor(JSONB())
        else:
            return dialect.type_descriptor(sa.JSON())


# the custom ser/de handlers are to allow the database to ensure
# uniqueness of dependency bindings (see https://github.com/NREL/BuildingMOTIF/pull/113)
def _custom_json_serializer(obj):
    """
    Serializes dictionaries as a sorted list of key-value tuples. All
    other items are serialized as normal
    """
    if isinstance(obj, dict):
        # ensure dictionary has at least 1 pair
        if len(obj) == 0:
            obj[None] = None
        return json.dumps(sorted([(k, v) for k, v in obj.items()]))
    return json.dumps(obj)


def _custom_json_deserializer(inp):
    """
    Handles deserializing the objects serialied by _custom_json_serializer
    """
    obj = json.loads(inp)
    # return normal object if it is not a list
    if not isinstance(obj, list):
        return obj
    # if *all* of the items in the list are pairs,
    # then we deserialize as a dictionary
    if len(obj) > 0 and all(map(lambda x: isinstance(x, list) and len(x) == 2, obj)):
        return dict(obj)
    # ...otherwise return. It's just a normal list!
    return obj



"""
Using the VF2 algorithm to compute subgraph isomorphisms between a template T and a graph G.
If the found isomorphism is a subgraph of T, then T is not fully matched and additional
input is required to fully populate the template.
"""
from collections import defaultdict
from itertools import combinations, permutations
from typing import TYPE_CHECKING, Callable, Dict, Generator, List, Optional, Set, Tuple

import networkx as nx  # type: ignore
from networkx.algorithms.isomorphism import DiGraphMatcher  # type: ignore
from rdflib import Graph, URIRef
from rdflib.extras.external_graph_libs import rdflib_to_networkx_digraph
from rdflib.term import Node

from buildingmotif.namespaces import OWL, PARAM, RDF, RDFS
from buildingmotif.utils import copy_graph

if TYPE_CHECKING:
    from buildingmotif.dataclasses.template import Template

Mapping = Dict[Node, Node]


# used to accelerate monomorphism search
# outer key is the address of the ontology graph
# inner map is from a class to its parent classes
class _ontology_lookup_cache:
    sc_cache: Dict[int, Dict[Node, Set[Node]]]
    t_cache: Dict[int, Dict[Node, Set[Node]]]
    in_cache: Dict[int, Dict[Node, bool]]

    def __init__(self):
        self.sc_cache = {}
        self.sp_cache = {}
        self.t_cache = {}
        self.in_cache = {}

    def parents(self, ntype: Node, ontology: Graph) -> Set[Node]:
        if id(ontology) not in self.sc_cache:
            self.sc_cache[id(ontology)] = {}
        cache = self.sc_cache[id(ontology)]
        # populate cache if necessary
        if ntype not in cache:
            cache[ntype] = set(ontology.transitive_objects(ntype, RDFS.subClassOf))  # type: ignore
        return cache[ntype]

    def superproperties(self, ntype: Node, ontology: Graph) -> Set[Node]:
        if id(ontology) not in self.sp_cache:
            self.sp_cache[id(ontology)] = {}
        cache = self.sp_cache[id(ontology)]
        # populate cache if necessary
        if ntype not in cache:
            cache[ntype] = set(ontology.transitive_objects(ntype, RDFS.subPropertyOf))
        return cache[ntype]

    def types(self, node: Node, graph: Graph) -> Set[URIRef]:
        if id(graph) not in self.t_cache:
            self.t_cache[id(graph)] = {}
        cache = self.t_cache[id(graph)]
        # populate cache if necessary
        if node not in cache:
            cache[node] = set(graph.objects(node, RDF.type))  # type: ignore
        if len(cache[node]) == 0:
            cache[node] = {OWL.NamedIndividual}
        return cache[node]  # type: ignore

    def defined_in(self, node: Node, graph: Graph) -> bool:
        if id(graph) not in self.in_cache:
            self.in_cache[id(graph)] = {}
        cache = self.in_cache[id(graph)]
        # populate cache if necessary
        if node not in cache:
            cache[node] = (node, RDF.type, OWL.Class) in graph
        return cache[node]


def _get_types(n: Node, g: Graph, _cache: _ontology_lookup_cache) -> Set[URIRef]:
    """
    Types for a node should only be URIRefs, so the
    type filtering here should be safe.
    Defaults to owl:NamedIndividual as a "root" type
    """
    return _cache.types(n, g)


def _compatible_types(
    n1: Node,
    g1: Graph,
    n2: Node,
    g2: Graph,
    ontology: Graph,
    _cache: _ontology_lookup_cache,
) -> bool:
    """
    Returns true if the two terms have covariant types.

    :param n1: First node
    :type n1: Node
    :param g1: The graph containing the first node's types
    :type g1: Graph
    :param n2: Second node
    :type n2: Node
    :param g2: The graph containing the second node's types
    :type g2: Graph
    :param ontology: The ontology graph that defines the class hierarchy
    :type ontology: Graph
    :return: True if the nodes are semantically compatible, false otherwise
    :rtype: bool
    """
    n1types = _get_types(n1, g1, _cache)
    n2types = _get_types(n2, g2, _cache)
    # check if these are properties; if so, use subPropertyOf, not subClassOf
    property_types = {OWL.ObjectProperty, OWL.DatatypeProperty}
    if property_types.intersection(n1types) and property_types.intersection(n2types):
        for n1type in _get_types(n1, g1, _cache):
            for n2type in _get_types(n2, g2, _cache):

                # check if types are covariant
                if n2type in _cache.superproperties(
                    n1type, ontology
                ) or n1type in _cache.superproperties(n2type, ontology):
                    return True
    else:
        for n1type in _get_types(n1, g1, _cache):
            for n2type in _get_types(n2, g2, _cache):

                # check if types are covariant
                if n2type in _cache.parents(
                    n1type, ontology
                ) or n1type in _cache.parents(n2type, ontology):
                    return True
    return False


def get_semantic_feasibility(
    G1: Graph, G2: Graph, ontology: Graph, _cache: _ontology_lookup_cache
) -> Callable[[Node, Node], bool]:
    """Returns a function that checks if two nodes are semantically feasible to
    be matched given the information in the provided ontology.

    The function returns true if the two nodes are semantically feasible to be
    matched. We use the following checks:
    1. If the two nodes are both classes, one must be a subclass of the other.
    2. If the two nodes are instances, they must be of the same class.

    :param G1: graph 1
    :type G1: Graph
    :param G2: graph 2
    :type G2: Graph
    :param ontology: ontology graph
    :type ontology: Graph
    :param _cache: ontology lookup cache
    :type _cache: _ontology_lookup_cache
    :return: function that checks if two nodes are semantically feasible
    :rtype: Callable[[Node, Node], bool]
    """
    # TODO: other checks?
    def semantic_feasibility(n1: Node, n2: Node) -> bool:
        """Checks if two nodes are semantically feasible to be matched.

        :param n1: node 1
        :type n1: Node
        :param n2: node 2
        :type n2: Node
        :return: true if nodes are feasible, false if not
        :rtype: bool
        """
        # case 0: same node
        if n1 == n2:
            return True
        # case 1: both are classes
        if _cache.defined_in(n1, ontology) and _cache.defined_in(n2, ontology):
            if n2 in _cache.parents(n1, ontology):
                return True
            elif n1 in _cache.parents(n2, ontology):
                return True
            else:
                return False
        # case 2: both are instances
        if _compatible_types(n1, G1, n2, G2, ontology, _cache):
            return True
        return False

    return semantic_feasibility


class _VF2SemanticMatcher(DiGraphMatcher):
    """
    A subclass of DiGraphMatcher that incorporates semantic feasibility into the matching
    process using the provided ontology.

    :param T: template graph
    :param G: building graph
    :param ontology: ontology that contains the information about node semantics
    """

    def __init__(self, T: Graph, G: Graph, ontology: Graph):
        super().__init__(rdflib_to_networkx_digraph(T), rdflib_to_networkx_digraph(G))
        self.ontology = ontology
        self._cache = _ontology_lookup_cache()
        self._semantic_feasibility = get_semantic_feasibility(
            T, G, ontology, self._cache
        )

    def semantic_feasibility(self, g1: Node, g2: Node) -> bool:
        """
        Returns true if the two nodes are semantically feasible to be matched.
        """
        val = self._semantic_feasibility(g1, g2)
        return val


def generate_all_subgraphs(T: Graph) -> Generator[Graph, None, None]:
    """Generates all node-induced subgraphs of T in order of decreasing size.

    We generate subgraphs in decreasing order of size because we want to find
    the largest subgraph as part of the monomorphism search process.

    :param T: template graph
    :type T: Graph
    :yield: subgraphs
    :rtype: Generator[Graph, None, None]
    """
    # no monomorphism will be larger than the number of distinct nodes in the graph
    largest_sg_size = len(T.all_nodes())
    for nodecount in range(largest_sg_size, 1, -1):
        for nodelist in combinations(T.all_nodes(), nodecount):
            yield digraph_to_rdflib(rdflib_to_networkx_digraph(T).subgraph(nodelist))


def digraph_to_rdflib(digraph: nx.DiGraph) -> Graph:
    """Turns a `nx.DiGraph` into an `rdflib.Graph`.

    Expects the nx.DiGraph to have been produced by
    :py:func:`rdflib_to_networkx_digraph`.

    :param digraph: directed graph
    :type digraph: nx.DiGraph
    :return: RDF graph
    :rtype: Graph
    """
    g = Graph()
    for s, o, pdict in digraph.edges(data=True):
        g.add((s, pdict["triples"][0][1], o))
    return g


class TemplateMatcher:
    """Computes the set of subgraphs of G that are monomorphic to T; these are
    organized by how "complete" the monomorphism is.
    """

    mappings: Dict[int, List[Mapping]]
    template: "Template"
    building: Graph
    template_bindings: Dict[str, Node]
    template_graph: Graph

    def __init__(
        self,
        building: Graph,
        template: "Template",
        ontology: Graph,
        graph_target: Optional[Node] = None,
    ):
        self.mappings = defaultdict(list)
        self.template_bindings = {}
        self.template = template
        self.building = building
        self.ontology = ontology
        self.graph_target = graph_target

        self.template_graph = copy_graph(template.body)
        self.template_parameters: Set[Node] = {
            PARAM[p] for p in self.template.parameters
        }

        self._generate_mappings()

    def _generate_mappings(self):
        for template_subgraph in generate_all_subgraphs(self.template_graph):
            matching = _VF2SemanticMatcher(
                self.building, template_subgraph, self.ontology
            )
            if matching.subgraph_is_monomorphic():
                for sg in matching.subgraph_monomorphisms_iter():
                    # skip if the subgraph does not contain the graph node we care about
                    if self.graph_target and self.graph_target not in sg.keys():
                        continue
                    # sg is a mapping from building graph nodes to template nodes
                    # that constitutes a monomorphism.
                    if len(sg) == 0:
                        continue
                    # TODO: Limit mappings to those that include all of the params?
                    # TODO: ignore optional parameters?
                    # TODO: require that 'name' is in the parameters?

                    # if there is an overlap, yield the subgraph
                    if set(sg.values()).intersection(self.template_parameters):
                        self.add_mapping(sg)

    def add_mapping(self, mapping: Mapping):
        """Adds a mapping to the set of mappings.

        :param mapping: mapping
        :type mapping: Mapping
        """
        if mapping not in self.mappings[len(mapping)]:
            self.mappings[len(mapping)].append(mapping)

    @property
    def largest_mapping_size(self) -> int:
        """Returns the size of the largest mapping.

        :return: size of largest mapping
        :rtype: int
        """
        return max(self.mappings.keys())

    def building_subgraph_from_mapping(self, mapping: Mapping) -> Graph:
        """Returns the subgraph of the building graph that corresponds to the
        given mapping.

        :param mapping: mapping
        :type mapping: Mapping
        :return: subgraph
        :rtype: Graph
        """
        g = rdflib_to_networkx_digraph(self.building)
        edges = permutations(mapping.keys(), 2)
        sg = g.edge_subgraph(edges)
        return digraph_to_rdflib(sg)

    def template_subgraph_from_mapping(self, mapping: Mapping) -> Graph:
        """Returns the subgraph of the template graph that corresponds to the
        given mapping.

        :param mapping: mapping
        :type mapping: Mapping
        :return: subgraph
        :rtype: Graph
        """
        # TODO: need to keep the edges that are more generic than what we have inside the graph.
        # For example, if the building has (x a brick:AHU) then we don't need to remind them to
        # add an edge (x a brick:Equipment) because that is redundant
        g = rdflib_to_networkx_digraph(self.template_graph)
        return digraph_to_rdflib(g.subgraph(mapping.values()))

    def remaining_template_graph(self, mapping: Mapping) -> Graph:
        """Returns the remaining template graph to be filled out given a
        mapping.

        :param mapping: mapping
        :type mapping: Mapping
        :return: remaining template graph to be filled out
        :rtype: Graph
        """
        sg = self.template_subgraph_from_mapping(mapping)
        return self.template_graph - sg

    def remaining_template(self, mapping: Mapping) -> Optional["Template"]:
        """Returns the remaining template to be filled out given a mapping.

        :param mapping: mapping
        :type mapping: Mapping
        :return: remaining template to be filled out
        :rtype: Optional[Template]
        """
        # if all parameters are fulfilled by the mapping, then return None
        mapping = {k: v for k, v in mapping.items() if str(v) in PARAM}
        mapped_params: Set[Node] = {v for v in mapping.values()}
        if not self.template_parameters - mapped_params:
            # return self.building_subgraph_from_mapping(mapping)
            return None
        bindings = {}
        for building_node, param in mapping.items():
            if param is not None:
                bindings[str(param)[len(PARAM) :]] = building_node
        # this *should* be a template because we don't have bindings for all of
        # the template's parameters
        res = self.template.evaluate(bindings)
        assert not isinstance(res, Graph)
        return res

    def mappings_iter(self, size=None) -> Generator[Mapping, None, None]:
        """Returns an iterator over all of the mappings of the given size.

        If size is None, then all mappings are returned in descending order
        of the size of the mapping. This means the most complete mappings
        will be returned first.

        :param size: size, defaults to None
        :type size: int, optional
        :yield: mapping iterator
        :rtype: Generator[Mapping, None, None]
        """
        if size is None:
            for size in sorted(self.mappings.keys(), reverse=True):
                for mapping in self.mappings[size]:
                    if not mapping:
                        continue
                    yield mapping
        else:
            for mapping in self.mappings[size]:
                if not mapping:
                    continue
                yield mapping

    def building_mapping_subgraphs_iter(
        self,
        size=None,
    ) -> Generator[Tuple[Mapping, Graph], None, None]:
        """Returns an iterator over all of the subgraphs with a mapping of the
        given size.

        If size is None, then all mappings are returned in descending order of
        the size of the mapping. This means the most complete subgraphs will be
        returned first.

        :param size: size, defaults to None
        :type size: int, optional
        :yield: mapping and subgraph iterator
        :rtype: Generator[Tuple[Mapping, Graph], None, None]
        """
        cache = set()
        for mapping in self.mappings_iter(size):
            subgraph = self.building_subgraph_from_mapping(mapping)
            if not subgraph.connected():
                continue
            key = tuple(sorted(subgraph.all_nodes()))  # type: ignore
            if key in cache:
                continue
            cache.add(key)
            yield mapping, subgraph



from typing import List, Optional, Tuple, Union

from rdflib import RDF, BNode, Graph, Literal, URIRef
from rdflib.collection import Collection
from rdflib.term import Node

from buildingmotif.namespaces import CONSTRAINT, SH, A, bind_prefixes


class Shape(Graph):
    """Base class for constructing shapes programatically"""

    def __init__(
        self,
        identifier: Optional[Union[Node, str]] = None,
        message: Optional[str] = None,
    ) -> None:
        """

        :param identifier: id for shape
        :type identifier: Optional[Union[Node, str]]
        :param message: sh:message annotation
        :type message: str
        """
        super().__init__(identifier=identifier)
        bind_prefixes(self)

        if message:
            self.add((self.identifier, SH["message"], Literal(message)))

    def add_property(self, property: URIRef, object: Node):
        """Add property to shape

        :param property: ref of property
        :type property: URIRef
        :param object: ref of object
        :type object: Node"""
        # This same functionality could be added with just use of add.
        # This design is useful to allow adding properties in the
        # builder paradigm.
        self.add((self, property, object))

        return self

    def add_list_property(
        self, property: URIRef, nodes: Union[List[Node], Tuple[Node, ...]]
    ):
        """Add property which references list to shape

        :param property: ref of property
        :type property: URIRef
        :param nodes: nodes to include in list
        :type nodes: Union[List[Node], Tuple[Node, ...]]"""
        identifier = BNode()
        Collection(self, identifier, nodes)

        self.add((self, property, identifier))

        return self

    def OR(self, *nodes: Node):
        """add OR property

        :param nodes: list of nodes to OR
        :type nodes: Union[List[Node], Tuple[Node, ...]]
        """
        self.add_list_property(SH["or"], nodes)
        return self

    def AND(self, *nodes: Node):
        """add AND property

        :param nodes: list of nodes to AND
        :type nodes: Union[List[Node], Tuple[Node, ...]]
        """
        self.add_list_property(SH["and"], nodes)
        return self

    def NOT(self, node: Node):
        """add NOT property

        :param nodes: list of nodes to NOT
        :type nodes: Union[List[Node], Tuple[Node, ...]]
        """
        self.add((self, SH["not"], node))

        return self

    def XONE(self, *nodes: Node):
        """add XONE property

        :param nodes: list of nodes to XONE
        :type nodes: Union[List[Node], Tuple[Node, ...]]
        """
        self.add_list_property(SH["xone"], nodes)
        return self

    def add(self, triple: Tuple[Node, Node, Node]):
        (s, p, o) = triple
        if isinstance(o, Shape):
            # if the object being added is of type Shape
            # add the whole graph to this graph
            self += o
            o = o.identifier
        if s == self:
            s = s.identifier
        triple = (s, p, o)
        return super().add(triple)


class NodeShape(Shape):
    """Class for constructing Node Shapes programatically"""

    def __init__(
        self,
        identifier: Optional[Union[Node, str]] = None,
        message: Optional[str] = None,
    ) -> None:
        """

        :param identifier: id for shape
        :type identifier: Optional[Union[Node, str]]
        :param message: sh:message annotation
        :type message: str
        """
        super().__init__(identifier=identifier, message=message)

        self.add((self, A, SH["NodeShape"]))

    def of_class(self, class_: Node, active=False):
        """Add constraint that target much be of a certain class

        :param class_: class of target
        :type class_: Node
        :param active: should shape actively target the class or not
        :type active: bool"""
        predicate = SH["targetClass"] if active else SH["class"]

        self.add((self, predicate, class_))
        self.add((self, CONSTRAINT["class"], class_))

        return self

    def always_run(self):
        """Add blank node target
        This target insures that the shape will always be evaluated.
        If the shape has properties this can cause it to fail."""
        self.add((self, SH["targetNode"], BNode()))
        return self

    def count(self, exactly: int = None):
        """Add an exact count constraint.

        :param exactly: exact number of instances of class to match
        :type exactly: int"""
        if exactly:
            self.add((self, CONSTRAINT["exactCount"], Literal(exactly)))
        return self

    def has_property(self, property: Union[Node, URIRef]):
        """Add property constraint.

        :param property: property shape or property to add constraint for
        :type property: Union[Node, URIRef]"""
        if isinstance(property, URIRef):
            property = PropertyShape().has_path(property)

        self.add((self, SH["property"], property))

        return self


class PropertyShape(Shape):
    def __init__(
        self,
        identifier: Optional[Union[Node, str]] = None,
        message: Optional[str] = None,
    ) -> None:
        """

        :param identifier: id for shape
        :type identifier: Optional[Union[Node, str]]
        :param message: sh:message annotation
        :type message: str
        """
        super().__init__(identifier=identifier, message=message)

        self.add((self, RDF.type, SH["PropertyShape"]))

    def has_path(
        self,
        path: Node,
        zero_or_one: bool = False,
        zero_or_more: bool = False,
        one_or_more: bool = False,
    ) -> "PropertyShape":
        """Add path constraint to shape.
        zero_or_one, zero_or_more, and one_or_more flags are mutually exclusive

        :param path: path to add constraint for
        :type path: Node
        :param zero_or_one: match zero or one instances of path
        :type zero_or_one: bool
        :param zero_or_more: match zero or more instances of path
        :type zero_or_more: bool
        :param one_or_more: match one or more instances of path
        :type one_or_more: bool"""

        if zero_or_one or zero_or_more or one_or_more:
            path_constraint = None
            if zero_or_one:
                path_constraint = SH["path-zero-or-one"]
            elif zero_or_more:
                path_constraint = SH["path-zero-or-more"]
            elif one_or_more:
                path_constraint = SH["path-one-or-more"]
            self.add((self, SH["path"], Shape().add_property(path_constraint, path)))
        else:
            self.add((self, SH["path"], path))

        return self

    def matches(
        self,
        target: Node,
        type: URIRef,
        min: int = None,
        max: int = None,
        exactly: int = None,
        qualified: bool = False,
    ):
        """Add target matches constraint to property shape

        :param target: target node to specify what should be matched, usually shape or class
        :type target: Node
        :param type: sh:class or sh:node
        :type type: URIRef
        :param min: min count of matched entities
        :type min: int
        :param max: max count of matched entities
        :type max: int
        :param exactly: exact count of matched entities (takes precidence over min/max)
        :type exactly: int
        :param qualified: Is this property qualified or universal
        :type qualified: bool

        """
        if min is None and max is None and exactly is None:
            if qualified:
                raise ValueError("min, max or exactly must have a value")
            else:
                self.add((self, type, target))
            return self

        if exactly is not None:
            min = max = exactly

        if qualified:
            blank_node = BNode()
            self.add((blank_node, type, target))
            self.add((self, SH["qualifiedValueShape"], blank_node))
            if min is not None:
                self.add((self, SH["qualifiedMinCount"], Literal(min)))
            if max is not None:
                self.add((self, SH["qualifiedMaxCount"], Literal(max)))
        else:
            self.add((self, type, target))
            if min is not None:
                self.add((self, SH["minCount"], Literal(min)))
            if max is not None:
                self.add((self, SH["maxCount"], Literal(max)))
        return self

    def matches_class(
        self,
        class_: URIRef,
        min: int = None,
        max: int = None,
        exactly: int = None,
        qualified=False,
    ):
        """Add target matches class constraint to property shape

        :param class_: target class what should be matched
        :type class_: Node
        :param min: min count of matched entities
        :type min: int
        :param max: max count of matched entities
        :type max: int
        :param exactly: exact count of matched entities (takes precidence over min/max)
        :type exactly: int
        :param qualified: Is this property qualified or universal
        :type qualified: bool
        """
        return self.matches(class_, SH["class"], min, max, exactly, qualified)

    def matches_shape(
        self,
        shape: Node,
        min: int = None,
        max: int = None,
        exactly: int = None,
        qualified=False,
    ):
        """Add target matches shape constraint to property shape

        :param shape: target shape what should be matched
        :type shape: Node
        :param min: min count of matched entities
        :type min: int
        :param max: max count of matched entities
        :type max: int
        :param exactly: exact count of matched entities (takes precidence over min/max)
        :type exactly: int
        :param qualified: Is this property qualified or universal
        :type qualified: bool
        """
        return self.matches(shape, SH["node"], min, max, exactly, qualified)


def OR(*nodes: Node) -> Shape:
    """add OR property

    :param nodes: list of nodes to OR
    :type nodes: Union[List[Node], Tuple[Node, ...]]
    """
    return Shape().OR(*nodes)


def AND(*nodes: Node) -> Shape:
    """add AND property

    :param nodes: list of nodes to AND
    :type nodes: Union[List[Node], Tuple[Node, ...]]
    """
    return Shape().AND(*nodes)


def NOT(node: Node) -> Shape:
    """add NOT property

    :param nodes: list of nodes to NOT
    :type nodes: Union[List[Node], Tuple[Node, ...]]
    """
    return Shape().NOT(node)


def XONE(*nodes: Node) -> Shape:
    """add XONE property

    :param nodes: list of nodes to XONE
    :type nodes: Union[List[Node], Tuple[Node, ...]]
    """
    return Shape().XONE(*nodes)



import logging
import pathlib
import tempfile
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Union

import pygit2
import rdflib
import yaml
from pkg_resources import resource_exists, resource_filename
from rdflib.exceptions import ParserError
from rdflib.plugins.parsers.notation3 import BadSyntax
from rdflib.util import guess_format

from buildingmotif import get_building_motif
from buildingmotif.database.errors import LibraryNotFound
from buildingmotif.database.tables import DBLibrary, DBTemplate
from buildingmotif.dataclasses.shape_collection import ShapeCollection
from buildingmotif.dataclasses.template import Template
from buildingmotif.schemas import validate_libraries_yaml
from buildingmotif.template_compilation import compile_template_spec
from buildingmotif.utils import get_ontology_files, shacl_inference, skip_uri

if TYPE_CHECKING:
    from buildingmotif import BuildingMOTIF


@dataclass
class _template_dependency:
    """Represents early-bound (template_id) or late-bound (template_name and
    library) dependency of a template on another template.
    """

    template_name: str
    bindings: Dict[str, Any]
    library: str
    template_id: Optional[int] = None

    def __repr__(self):
        return (
            f"dep<name={self.template_name} bindings={self.bindings} "
            f"library={self.library} id={self.template_id}>"
        )

    @classmethod
    def from_dict(
        cls, d: Dict[str, Any], dependent_library_name: str
    ) -> "_template_dependency":
        """Creates a py:class:`_template_dependency` from a dictionary.

        :param d: dictionary
        :type d: Dict[str, Any]
        :param dependent_library_name: library name
        :type dependent_library_name: str
        :return: the _template_dependency from the dict
        :rtype: _template_dependency
        """
        template_name = d["template"]
        bindings = d.get("args", {})
        library = d.get("library", dependent_library_name)
        template_id = d.get("template_id")
        return cls(template_name, bindings, library, template_id)

    def to_template(self, id_lookup: Dict[str, int]) -> Template:
        """Resolve this dependency to a template.

        :param id_lookup: a local cache of {name: id} for uncommitted templates
        :type id_lookup: Dict[str, int]
        :return: the template instance this dependency points to
        :rtype: Template
        """
        # direct lookup if id is provided
        if self.template_id is not None:
            return Template.load(self.template_id)
        # if id is not provided, look at our local 'cache' of to-be-committed
        # templates for the id (id_lookup)
        if self.template_name in id_lookup:
            return Template.load(id_lookup[self.template_name])
        # if not in the local cache, then search the database for the template
        # within the given library
        library = Library.load(name=self.library)
        return library.get_template_by_name(self.template_name)


@dataclass
class Library:
    """This class mirrors :py:class:`database.tables.DBLibrary`."""

    _id: int
    _name: str
    _bm: "BuildingMOTIF"

    @classmethod
    def create(cls, name: str, overwrite: Optional[bool] = True) -> "Library":
        """Create new Library.

        :param name: library name
        :type name: str
        :param overwrite: if True, overwrite the existing copy of the library.
        :type overwrite: Optional[bool]
        :return: new library
        :rtype: Library
        """
        bm = get_building_motif()
        try:
            db_library = bm.table_connection.get_db_library_by_name(name)
            if overwrite:
                cls._clear_library(db_library)
            else:
                logging.warning(
                    f'Library {name} already exists in database. To ovewrite load library with "overwrite=True"'  # noqa
                )
        except LibraryNotFound:
            db_library = bm.table_connection.create_db_library(name)

        return cls(_id=db_library.id, _name=db_library.name, _bm=bm)

    @classmethod
    def _clear_library(cls, library: DBLibrary) -> None:
        """Clear contents of a library.

        :param library: library to clear
        :type library: DBLibrary
        """
        bm = get_building_motif()
        for template in library.templates:  # type: ignore
            bm.session.delete(template)

    # TODO: load library from URI? Does the URI identify the library uniquely?
    # TODO: can we deduplicate shape graphs? use hash of graph?
    @classmethod
    def load(
        cls,
        db_id: Optional[int] = None,
        ontology_graph: Optional[Union[str, rdflib.Graph]] = None,
        directory: Optional[str] = None,
        name: Optional[str] = None,
        overwrite: Optional[bool] = True,
        infer_templates: Optional[bool] = True,
        run_shacl_inference: Optional[bool] = True,
    ) -> "Library":
        """Loads a library from the database or an external source.
        When specifying a path to load a library or ontology_graph from,
        paths within the buildingmotif.libraries module will be prioritized
        if they resolve.

        :param db_id: the unique id of the library in the database,
            defaults to None
        :type db_id: Optional[int], optional
        :param ontology_graph: a path to a serialized RDF graph.
            Supports remote ontology URLs, defaults to None
        :type ontology_graph: Optional[str|rdflib.Graph], optional
        :param directory: a path to a directory containing a library,
            or an rdflib graph, defaults to None
        :type directory: Optional[str], optional
        :param name: the name of the library inside the database,
            defaults to None
        :type name: Optional[str], optional
        :param overwrite: if true, replace any existing copy of the
            library, defaults to True
        :type overwrite: Optional[true], optional
        :param infer_templates: if true, infer shapes from the ontology graph,
            defaults to True
        :type infer_templates: Optional[bool], optional
        :param run_shacl_inference: if true, run SHACL inference on the ontology graph,
            using the BuildingMOTIF SHACL engine, defaults to True
        :type run_shacl_inference: Optional[bool], optional
        :return: the loaded library
        :rtype: Library
        :raises Exception: if the library cannot be loaded
        """
        if db_id is not None:
            return cls._load_from_db(db_id)
        elif ontology_graph is not None:
            if isinstance(ontology_graph, str):
                ontology_graph_path = ontology_graph
                if resource_exists("buildingmotif.libraries", ontology_graph_path):
                    logging.debug(f"Loading builtin library: {ontology_graph_path}")
                    ontology_graph_path = resource_filename(
                        "buildingmotif.libraries", ontology_graph_path
                    )
                ontology_graph = rdflib.Graph()
                ontology_graph.parse(
                    ontology_graph_path, format=guess_format(ontology_graph_path)
                )
            return cls._load_from_ontology(
                ontology_graph,
                overwrite=overwrite,
                infer_templates=infer_templates,
                run_shacl_inference=run_shacl_inference,
            )
        elif directory is not None:
            if resource_exists("buildingmotif.libraries", directory):
                logging.debug(f"Loading builtin library: {directory}")
                src = pathlib.Path(
                    resource_filename("buildingmotif.libraries", directory)
                )
            else:
                src = pathlib.Path(directory)
            if not src.exists():
                raise Exception(f"Directory {src} does not exist")
            return cls._load_from_directory(
                src,
                overwrite=overwrite,
                infer_templates=infer_templates,
                run_shacl_inference=run_shacl_inference,
            )
        elif name is not None:
            bm = get_building_motif()
            db_library = bm.table_connection.get_db_library_by_name(name)
            return cls(_id=db_library.id, _name=db_library.name, _bm=bm)
        else:
            raise Exception("No library information provided")

    @classmethod
    def _load_from_db(cls, id: int) -> "Library":
        """Load library from database by id.

        :param id: id of library
        :type id: int
        :return: library
        :rtype: Library
        """
        bm = get_building_motif()
        db_library = bm.table_connection.get_db_library(id)

        return cls(_id=db_library.id, _name=db_library.name, _bm=bm)

    @classmethod
    def _load_from_ontology(
        cls,
        ontology: rdflib.Graph,
        overwrite: Optional[bool] = True,
        infer_templates: Optional[bool] = True,
        run_shacl_inference: Optional[bool] = True,
    ) -> "Library":
        """
        Load a library from an ontology graph. This proceeds as follows.
        First, get all entities in the graph that are instances of *both* owl:Class
        and sh:NodeShape. (this is "candidates")

        For each candidate, use the utility function to parse the NodeShape and turn
        it into a Template.

        :param ontology: the graph to load into BuildingMOTIF and interpret as a Library
        :type ontology: rdflib.Graph
        :param overwrite: if true, overwrite the existing copy of the Library
        :type overwrite: bool
        :param infer_templates: if true, infer shapes from the ontology graph
        :type infer_templates: bool
        :param run_shacl_inference: if true, run SHACL inference on the ontology graph
        :type run_shacl_inference: bool
        :return: the loaded Library
        :rtype: "Library"
        """
        # get the name of the ontology; this will be the name of the library
        # any=False will raise an error if there is more than one ontology defined  in the graph
        ontology_name = ontology.value(
            predicate=rdflib.RDF.type, object=rdflib.OWL.Ontology, any=False
        ) or rdflib.URIRef("urn:unnamed/")

        if not overwrite:
            if cls._library_exists(ontology_name):
                logging.warning(
                    f'Library "{ontology_name}" already exists in database and "overwrite=False". Returning existing library.'  # noqa
                )
                return Library.load(name=ontology_name)

        # expand the ontology graph before we insert it into the database. This will ensure
        # that the output of compiled models will not contain triples that really belong to
        # the ontology
        if run_shacl_inference:
            ontology = shacl_inference(
                ontology, engine=get_building_motif().shacl_engine
            )

        lib = cls.create(ontology_name, overwrite=overwrite)

        # load the ontology graph as a shape_collection
        shape_col_id = lib.get_shape_collection().id
        assert shape_col_id is not None  # should always pass
        shape_col = ShapeCollection.load(shape_col_id)
        shape_col.add_graph(ontology)

        if infer_templates:
            # infer shapes from any class/nodeshape candidates in the graph
            shape_col.infer_templates(lib)

        return lib

    def _load_shapes_from_directory(
        self,
        directory: pathlib.Path,
        infer_templates: Optional[bool] = True,
        run_shacl_inference: Optional[bool] = True,
    ):
        """Helper method to read all graphs in the given directory into this
        library.

        :param directory: directory containing graph files
        :type directory: pathlib.Path
        :param infer_templates: if true, infer shapes from the ontology graph
        :type infer_templates: bool
        :param run_shacl_inference: if true, run SHACL inference on the ontology graph
        :type run_shacl_inference: bool
        """
        shape_col_id = self.get_shape_collection().id
        assert shape_col_id is not None  # this should always pass
        shape_col = ShapeCollection.load(shape_col_id)
        for filename in get_ontology_files(directory):
            try:
                shape_col.graph.parse(filename, format=guess_format(filename))
            except (ParserError, BadSyntax) as e:
                logging.getLogger(__name__).error(
                    f"Could not parse file {filename}: {e}"
                )
                raise e
        if run_shacl_inference:
            shape_col.graph = shacl_inference(
                shape_col.graph, engine=get_building_motif().shacl_engine
            )
        # infer shapes from any class/nodeshape candidates in the graph
        if infer_templates:
            shape_col.infer_templates(self)

    @classmethod
    def _load_from_directory(
        cls,
        directory: pathlib.Path,
        overwrite: Optional[bool] = True,
        infer_templates: Optional[bool] = True,
        run_shacl_inference: Optional[bool] = True,
    ) -> "Library":
        """
        Load a library from a directory.

        Templates are read from YML files in the directory. The name of the
        library is given by the name of the directory.

        :param directory: directory containing a library
        :type directory: pathlib.Path
        :param overwrite: if true, overwrite the existing copy of the Library
        :type overwrite: bool
        :param infer_templates: if true, infer shapes from the ontology graph
        :type infer_templates: bool
        :param run_shacl_inference: if true, run SHACL inference on the ontology graph
        :type run_shacl_inference: bool
        :raises e: if cannot create template
        :raises e: if cannot resolve dependencies
        :return: library
        :rtype: Library
        """

        if not overwrite:
            if cls._library_exists(directory.name):
                logging.warning(
                    f'Library "{directory.name}" already exists in database and "overwrite=False". Returning existing library.'  # noqa
                )
                return Library.load(name=directory.name)

        lib = cls.create(directory.name, overwrite=overwrite)

        # setup caches for reading templates
        template_id_lookup: Dict[str, int] = {}
        dependency_cache: Dict[int, List[_template_dependency]] = {}

        # read all .yml files
        for file in directory.rglob("*.yml"):
            # if .ipynb_checkpoints, skip; these are cached files that Jupyter creates
            if ".ipynb_checkpoints" in file.parts:
                continue
            lib._read_yml_file(file, template_id_lookup, dependency_cache)
        # now that we have all the templates, we can populate the dependencies
        lib._resolve_template_dependencies(template_id_lookup, dependency_cache)
        # load shape collections from all ontology files in the directory
        lib._load_shapes_from_directory(directory)

        return lib

    @classmethod
    def load_from_libraries_yml(cls, filename: str):
        """
        Loads *multiple* libraries from a properly-formatted 'libraries.yml'
        file. Does not return a Library! You will need to load the libraries by
        name in order to get the dataclasses.Library object. We recommend loading
        libraries directly, one-by-one, in most cases. This method is here to support
        the commandline tool.

        :param filename: the filename of the YAML file to load library names from
        :type filename: str
        :rtype: None
        """
        libraries = yaml.load(open(filename, "r"), Loader=yaml.FullLoader)
        validate_libraries_yaml(libraries)  # raises exception
        for description in libraries:
            _resolve_library_definition(description)

    @staticmethod
    def _library_exists(library_name: str) -> bool:
        """Checks whether a library with the given name exists in the database."""
        bm = get_building_motif()
        try:
            bm.table_connection.get_db_library_by_name(library_name)
            return True
        except LibraryNotFound:
            return False

    def _resolve_dependency(
        self,
        template: Template,
        dep: Union[_template_dependency, dict],
        template_id_lookup: Dict[str, int],
    ):
        """Resolve a dependency to a template.

        :param template: template to resolve dependency for
        :type template: Template
        :param dep: dependency
        :type dep: Union[_template_dependency, dict]
        :param template_id_lookup: a local cache of {name: id} for uncommitted templates
        :type template_id_lookup: Dict[str, int]
        :return: the template instance this dependency points to
        :rtype: Template
        """
        # if dep is a _template_dependency, turn it into a template
        if isinstance(dep, _template_dependency):
            dependee = dep.to_template(template_id_lookup)
            template.add_dependency(dependee, dep.bindings)
            return

        # now, we know that dep is a dict

        # if dependency names a library explicitly, load that library and get the template by name
        if "library" in dep:
            dependee = Library.load(name=dep["library"]).get_template_by_name(
                dep["template"]
            )
            template.add_dependency(dependee, dep["args"])
            return
        # if no library is provided, try to resolve the dependency from this library
        if dep["template"] in template_id_lookup:
            dependee = Template.load(template_id_lookup[dep["template"]])
            template.add_dependency(dependee, dep["args"])
            return
        # check documentation for skip_uri for what URIs get skipped
        if skip_uri(dep["template"]):
            return

        # if the dependency is not in the local cache, then search through this library's imports
        # for the template
        for imp in self.graph_imports:
            try:
                library = Library.load(name=str(imp))
                dependee = library.get_template_by_name(dep["template"])
                template.add_dependency(dependee, dep["args"])
                return
            except Exception as e:
                logging.debug(
                    f"Could not find dependee {dep['template']} in library {imp}: {e}"
                )
        logging.warning(
            f"Warning: could not find dependee {dep['template']} in libraries {self.graph_imports}"
        )

    def _resolve_template_dependencies(
        self,
        template_id_lookup: Dict[str, int],
        dependency_cache: Mapping[int, Union[List[_template_dependency], List[dict]]],
    ):
        """Resolve all dependencies for all templates in this library"""
        # two phases here: first, add all of the templates and their dependencies
        # to the database but *don't* check that the dependencies are valid yet
        for template in self.get_templates():
            if template.id not in dependency_cache:
                continue
            for dep in dependency_cache[template.id]:
                self._resolve_dependency(template, dep, template_id_lookup)
        # check that all dependencies are valid (use parameters that exist, etc)
        for template in self.get_templates():
            template.check_dependencies()

    def _read_yml_file(
        self,
        file: pathlib.Path,
        template_id_lookup: Dict[str, int],
        dependency_cache: Dict[int, List[_template_dependency]],
    ):
        """Read a YML file into this library. Utility function for `_load_from_directory`."""
        contents = yaml.load(open(file, "r"), Loader=yaml.FullLoader)
        for templ_name, templ_spec in contents.items():
            # compile the template body using its rules
            templ_spec = compile_template_spec(templ_spec)
            # input name of template
            templ_spec.update({"name": templ_name})
            # remove dependencies so we can resolve them to their IDs later
            deps = [
                _template_dependency.from_dict(d, self.name)
                for d in templ_spec.pop("dependencies", [])
            ]
            templ_spec["optional_args"] = templ_spec.pop("optional", [])
            try:
                templ = self.create_template(**templ_spec)
            except Exception as e:
                logging.error(
                    f"Error creating template {templ_name} from file {file}: {e}"
                )
                raise e
            dependency_cache[templ.id] = deps
            template_id_lookup[templ.name] = templ.id

    @property
    def id(self) -> Optional[int]:
        return self._id

    @id.setter
    def id(self, new_id):
        raise AttributeError("Cannot modify db id")

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._bm.table_connection.update_db_library_name(self._id, new_name)
        self._name = new_name

    @property
    def graph_imports(self) -> List[rdflib.URIRef]:
        """
        Get the list of owl:imports for this library's shape collection
        """
        shape_col = self.get_shape_collection()
        return [
            i
            for i in shape_col.graph.objects(None, rdflib.OWL.imports)
            if isinstance(i, rdflib.URIRef)
        ]

    def create_template(
        self,
        name: str,
        body: Optional[rdflib.Graph] = None,
        optional_args: Optional[List[str]] = None,
    ) -> Template:
        """Create template in this library.

        :param name: name
        :type name: str
        :param body: template body
        :type body: rdflib.Graph
        :param optional_args: optional parameters for the template
        :type optional_args: list[str]
        :return: created template
        :rtype: Template
        """
        db_template = self._bm.table_connection.create_db_template(name, self._id)
        body = self._bm.graph_connection.create_graph(
            db_template.body_id, body if body else rdflib.Graph()
        )
        # ensure the "param" namespace is bound to the graph
        body.namespace_manager = self._bm.template_ns_mgr
        if optional_args is None:
            optional_args = []
        self._bm.table_connection.update_db_template_optional_args(
            db_template.id, optional_args
        )

        return Template(
            _id=db_template.id,
            _name=db_template.name,
            body=body,
            optional_args=optional_args,
            _bm=self._bm,
        )

    def get_templates(self) -> List[Template]:
        """Get templates from library.

        :return: list of templates
        :rtype: List[Template]
        """
        db_library = self._bm.table_connection.get_db_library(self._id)
        templates: List[DBTemplate] = db_library.templates
        return [Template.load(t.id) for t in templates]

    def get_shape_collection(self) -> ShapeCollection:
        """Get ShapeCollection from library.

        :return: library's shape collection
        :rtype: ShapeCollection
        """
        # TODO: we should save the libraries shape_collection to a class attr on load/create. That
        # way we wont need an additional db query each time we call this function.
        db_library = self._bm.table_connection.get_db_library(self._id)

        return ShapeCollection.load(db_library.shape_collection.id)

    def get_template_by_name(self, name: str) -> Template:
        """Get template by name from library.

        :param name: template name
        :type name: str
        :raises ValueError: if template not in library
        :return: template
        :rtype: Template
        """
        dbt = self._bm.table_connection.get_db_template_by_name(name)
        if dbt.library_id != self._id:
            raise ValueError(f"Template {name} not in library {self._name}")
        return Template.load(dbt.id)


def _resolve_library_definition(desc: Dict[str, Any]):
    """
    Loads a library from a description in libraries.yml
    """
    if "directory" in desc:
        spath = pathlib.Path(desc["directory"]).absolute()
        if spath.exists() and spath.is_dir():
            logging.info(f"Load local library {spath} (directory)")
            Library.load(directory=str(spath))
        else:
            raise Exception(f"{spath} is not an existing directory")
    elif "ontology" in desc:
        ont = desc["ontology"]
        g = rdflib.Graph().parse(ont, format=rdflib.util.guess_format(ont))
        logging.info(f"Load library {ont} as ontology graph")
        Library.load(ontology_graph=g)
    elif "git" in desc:
        repo = desc["git"]["repo"]
        branch = desc["git"]["branch"]
        path = desc["git"]["path"]
        logging.info(f"Load library {path} from git repository: {repo}@{branch}")
        with tempfile.TemporaryDirectory() as temp_loc:
            pygit2.clone_repository(
                repo, temp_loc, checkout_branch=branch
            )  # , depth=1)
            new_path = pathlib.Path(temp_loc) / pathlib.Path(path)
            if new_path.is_dir():
                _resolve_library_definition({"directory": new_path})
            else:
                _resolve_library_definition({"ontology": new_path})



from typing import Optional

# The point of these exceptions is to provide more specific error messages which
# also embed the ID or name of the object that was not found. This is useful for
# debugging and logging purposes.


class LibraryNotFound(Exception):
    def __init__(self, name: Optional[str] = None, idnum: Optional[int] = None):
        self.lib_name = name
        self.lib_id = idnum

    def __str__(self):
        if self.lib_name:
            return f"Name: {self.lib_name}"
        return f"ID: {self.lib_id}"


class ModelNotFound(Exception):
    def __init__(self, name: Optional[str] = None, idnum: Optional[int] = None):
        self.model_name = name
        self.model_id = idnum

    def __str__(self):
        if self.model_name:
            return f"Name: {self.model_name}"
        return f"ID: {self.model_id}"


class ShapeCollectionNotFound(Exception):
    def __init__(self, name: Optional[str] = None, idnum: Optional[int] = None):
        self.shape_collection_name = name
        self.shape_collection_id = idnum

    def __str__(self):
        if self.shape_collection_name:
            return f"Name: {self.shape_collection_name}"
        return f"ID: {self.shape_collection_id}"


class TemplateNotFound(Exception):
    def __init__(self, name: Optional[str] = None, idnum: Optional[int] = None):
        self.template_name = name
        self.template_id = idnum

    def __str__(self):
        if self.template_name:
            return f"Name: {self.template_name}"
        return f"ID: {self.template_id}"



from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Optional

import pandas as pd
import rdflib
import rdflib.query
from rdflib import URIRef

from buildingmotif.dataclasses.model import Model
from buildingmotif.dataclasses.shape_collection import ShapeCollection
from buildingmotif.dataclasses.validation import ValidationContext
from buildingmotif.namespaces import OWL, SH, A
from buildingmotif.utils import (
    copy_graph,
    rewrite_shape_graph,
    shacl_inference,
    shacl_validate,
    skolemize_shapes,
)


@dataclass
class CompiledModel:
    """
    This class represents a model that has been compiled against a set of ShapeCollections.
    """

    model: Model
    shape_collections: List[ShapeCollection]
    _compiled_graph: rdflib.Graph

    def __init__(
        self,
        model: Model,
        shape_collections: List[ShapeCollection],
        compiled_graph: rdflib.Graph,
        shacl_engine: str = "default",
    ):
        self.model = model
        self.shape_collections = shape_collections
        ontology_graph = rdflib.Graph()
        for shape_collection in shape_collections:
            ontology_graph += shape_collection.graph

        ontology_graph = skolemize_shapes(ontology_graph)

        shacl_engine = (
            self.model._bm.shacl_engine
            if (shacl_engine == "default" or not shacl_engine)
            else shacl_engine
        )

        self._compiled_graph = shacl_inference(
            compiled_graph, ontology_graph, shacl_engine
        )

    @cached_property
    def graph(self) -> rdflib.Graph:
        g = copy_graph(self._compiled_graph)
        for shape_collection in self.shape_collections:
            g += shape_collection.graph
        return g

    def validate_model_against_shapes(
        self,
        shapes_to_test: List[rdflib.URIRef],
        target_class: rdflib.URIRef,
    ) -> Dict[rdflib.URIRef, "ValidationContext"]:
        """Validates the model against a list of shapes and generates a
        validation report for each.

        :param shapes_to_test: list of shape URIs to validate the model against
        :type shapes_to_test: List[URIRef]
        :param target_class: the class upon which to run the selected shapes
        :type target_class: URIRef
        :return: a dictionary that relates each shape to test URIRef to a
                 ValidationContext
        :rtype: Dict[URIRef, ValidationContext]
        """
        model_graph = copy_graph(self._compiled_graph)

        results = {}

        targets = model_graph.query(
            f"""
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            SELECT ?target
            WHERE {{
                ?target rdf:type/rdfs:subClassOf* <{target_class}>

            }}
        """
        )
        # skolemize the shape graph so we have consistent identifiers across
        # validation through the interpretation of the validation report
        ontology_graph = self.graph.skolemize()

        for shape_uri in shapes_to_test:
            temp_model_graph = copy_graph(model_graph)
            for (s,) in targets:
                temp_model_graph.add((URIRef(s), A, shape_uri))

            valid, report_g, report_str = shacl_validate(
                temp_model_graph, ontology_graph, engine=self.model._bm.shacl_engine
            )

            results[shape_uri] = ValidationContext(
                self.shape_collections,
                ontology_graph,
                valid,
                report_g,
                report_str,
                self.model,
            )

        return results

    def validate(
        self,
        error_on_missing_imports: bool = True,
    ) -> "ValidationContext":
        """Validates this model against the given list of ShapeCollections.
        If no list is provided, the model will be validated against the model's "manifest".
        If a list of shape collections is provided, the manifest will *not* be automatically
        included in the set of shape collections.

        Loads all of the ShapeCollections into a single graph.

        :param error_on_missing_imports: if True, raises an error if any of the dependency
            ontologies are missing (i.e. they need to be loaded into BuildingMOTIF), defaults
            to True
        :type error_on_missing_imports: bool, optional
        :return: An object containing useful properties/methods to deal with
            the validation results
        :rtype: ValidationContext
        """
        # TODO: determine the return types; At least a bool for valid/invalid,
        # but also want a report. Is this the base pySHACL report? Or a useful
        # transformation, like a list of deltas for potential fixes?
        shapeg = copy_graph(self._compiled_graph)
        # aggregate shape graphs
        for sc in self.shape_collections:
            shapeg += sc.resolve_imports(
                error_on_missing_imports=error_on_missing_imports
            ).graph
        # inline sh:node for interpretability
        shapeg = rewrite_shape_graph(shapeg)

        # remove imports from sg
        shapeg.remove((None, OWL.imports, None))

        # skolemize the shape graph so we have consistent identifiers across
        # validation through the interpretation of the validation report
        shapeg = skolemize_shapes(shapeg)

        # remove imports from data graph
        shapeg.remove((None, OWL.imports, None))

        # validate the data graph
        valid, report_g, report_str = shacl_validate(
            shapeg, engine=self.model._bm.shacl_engine
        )
        return ValidationContext(
            self.shape_collections,
            shapeg,
            valid,
            report_g,
            report_str,
            self.model,
        )

    def defining_shape_collection(
        self, shape: rdflib.URIRef
    ) -> Optional[ShapeCollection]:
        """
        Given a shape, return the ShapeCollection that defines it. The search is limited to the
        ShapeCollections that were used to compile this model.

        :param shape: the shape to search for
        :type shape: rdflib.URIRef
        :return: the ShapeCollection that defines the shape, or None if the shape is not defined
        :rtype: Optional[ShapeCollection]
        """
        for sc in self.shape_collections:
            if (shape, A, SH.NodeShape) in sc.graph:
                return sc
        return None

    def shape_to_table(self, shape: rdflib.URIRef, table: str, conn):
        """
        Turn the shape into a SPARQL query and execute it on the model's graph, storing the results in a table.

        :param shape: the shape to query
        :type shape: rdflib.URIRef
        :param table: the name of the table to store the results in
        :type table: str
        :param conn: the connection to the database
        :type conn: sqlalchemy.engine.base.Connection
        """
        metadata = self.shape_to_df(shape)
        metadata.to_sql(table, conn, if_exists="replace", index=False)

    def shape_to_df(self, shape: rdflib.URIRef) -> pd.DataFrame:
        """
        Turn the shape into a SPARQL query and execute it on the model's graph, storing the results in a dataframe.

        :param shape: the shape to query
        :type shape: rdflib.URIRef
        :return: the results of the query
        :rtype: pd.DataFrame
        """
        defining_sc = self.defining_shape_collection(shape)
        if defining_sc is None:
            raise ValueError(
                f"Shape {shape} is not defined in any of the shape collections"
            )
        query = defining_sc.shape_to_query(shape)
        metadata = pd.DataFrame(self.graph.query(query).bindings, dtype="string")
        # metadata.columns will be rdflib.term.Variable objects, so we need to convert them to strings
        metadata.columns = [str(col) for col in metadata.columns]
        # convert the rdflib terms to Python types
        return metadata.map(lambda x: x.toPython())



import json
from dataclasses import dataclass
from functools import cached_property
from os import PathLike
from pathlib import Path
from typing import List

from rdflib import Graph, Namespace

from buildingmotif import BuildingMOTIF


@dataclass
class Record:
    """Represents a piece of metadata from some metadata ingress"""

    # an arbitrary "type hint"
    rtype: str
    # possibly-nested dictionary of (semi-)structured data from
    # the underlying source
    fields: dict


class IngressHandler:
    """Abstract superclass for Record/Graph ingress handlers"""

    pass


class RecordIngressHandler(IngressHandler):
    """Generates Record instances from an underlying metadata source"""

    def __init__(self, bm: BuildingMOTIF):
        self.bm = bm

    @cached_property
    def records(self) -> List[Record]:
        """
        Generates (then caches) a list of Records from an underlying data source
        """
        raise NotImplementedError("Must be overridden by subclass")

    def dump(self, path: PathLike):
        """
        Takes the contents of the records of this handler and writes them to a JSON file

        :param path: path to write output file to
        :type path: PathLike
        """
        output_string = self.dumps()
        output_file = Path(path)
        with output_file.open("w", encoding="utf-8") as f:
            f.write(output_string)

    def dumps(self) -> str:
        """
        Takes the contents of the records of this handler and writes them to a string
        """
        records = [
            {"rtype": record.rtype, "fields": record.fields} for record in self.records
        ]
        return json.dumps(records)

    @classmethod
    def load(cls, path: PathLike):
        """
        Takes a file generated by 'dump' and creates a new ingress handler with those records
        """
        return cls.loads(Path(path).read_text())

    @classmethod
    def loads(cls, s: str):
        """
        Takes the string output by 'dumps' and creates a new ingress handler with those records
        """
        self = cls.__new__(cls)
        records = []
        for record in json.loads(s):
            records.append(Record(record["rtype"], record["fields"]))
        self.records = records
        return self


class GraphIngressHandler(IngressHandler):
    """Generates a Graph from an underlying metadata source or RecordIngressHandler"""

    def __init__(self, bm: BuildingMOTIF):
        self.bm = bm

    def graph(self, ns: Namespace) -> Graph:
        """Generates an RDF graph with all entities being placed in the given namespace"""
        raise NotImplementedError("Must be overridden by subclass")



import flask
from flask import Blueprint, current_app, jsonify, request
from flask_api import status
from rdflib import Graph, URIRef
from rdflib.plugins.parsers.notation3 import BadSyntax

from buildingmotif.api.serializers.model import serialize
from buildingmotif.database.errors import (
    LibraryNotFound,
    ModelNotFound,
    ShapeCollectionNotFound,
)
from buildingmotif.dataclasses import Library, Model, ShapeCollection

blueprint = Blueprint("models", __name__)


@blueprint.route("", methods=(["GET"]))
def get_all_models() -> flask.Response:
    """Get all models.

    :return: all models
    :rtype: flask.Response
    """
    db_models = current_app.building_motif.table_connection.get_all_db_models()

    return jsonify(serialize(db_models))


@blueprint.route("/<models_id>", methods=(["GET"]))
def get_model(models_id: int) -> flask.Response:
    """Get Model by id.

    :param models_id: model id
    :type models_id: int
    :return: requested model
    :rtype: flask.Response
    """
    try:
        model = current_app.building_motif.table_connection.get_db_model(models_id)
    except ModelNotFound:
        return {"message": f"ID: {models_id}"}, status.HTTP_404_NOT_FOUND

    return jsonify(serialize(model)), status.HTTP_200_OK


@blueprint.route("/<models_id>/graph", methods=(["GET"]))
def get_model_graph(models_id: int) -> Graph:
    """Get model graph by id.

    :param models_id: model id
    :type models_id: int
    :return: requested model graph
    :rtype: rdflib.Graph
    """
    try:
        model = Model.load(models_id)
    except ModelNotFound:
        return {"message": f"ID: {models_id}"}, status.HTTP_404_NOT_FOUND

    g = Graph() + model.graph

    return g.serialize(format="ttl"), status.HTTP_200_OK


@blueprint.route("/<models_id>/target_nodes", methods=(["GET"]))
def get_target_nodes(models_id: int) -> Graph:
    """Get model graph by id.

    :param models_id: model id
    :type models_id: int
    :return: requested model graph
    :rtype: rdflib.Graph
    """
    try:
        model = Model.load(models_id)
    except ModelNotFound:
        return {"message": f"ID: {models_id}"}, status.HTTP_404_NOT_FOUND

    result = model.graph.query(
        """
        SELECT ?type WHERE {
        ?target rdf:type ?type
        }
    """
    )
    result = list({r for r in result})
    result.sort()

    return result, status.HTTP_200_OK


@blueprint.route("", methods=(["POST"]))
def create_model() -> flask.Response:
    """Create model

    :return: new model
    :rtype: Model
    """
    if request.content_type != "application/json":
        return {
            "message": "request content type must be json"
        }, status.HTTP_400_BAD_REQUEST

    name = request.json.get("name")
    description = request.json.get("description")

    if name is None:
        return {"message": "must give name"}, status.HTTP_400_BAD_REQUEST

    try:
        model = Model.create(name, description)
    except ValueError:
        return {
            "message": f"{name} does not look like a valid URI, "
            "trying to serialize this will break."
        }, status.HTTP_400_BAD_REQUEST

    current_app.building_motif.session.commit()

    model = current_app.building_motif.table_connection.get_db_model(model.id)

    return jsonify(serialize(model)), status.HTTP_201_CREATED


@blueprint.route("/<models_id>/graph", methods=(["PATCH", "PUT"]))
def update_model_graph(models_id: int) -> flask.Response:
    """Update model graph.

    Takes xml body of ttl formated graph.

    :param models_id: model id
    :type models_id: int
    :return: updated model graph
    :rtype: rdflib.Graph
    """
    try:
        model = Model.load(models_id)
    except ModelNotFound:
        return {"message": f"ID: {models_id}"}, status.HTTP_404_NOT_FOUND

    if request.content_type != "application/xml":
        return {
            "message": "request content type must be xml"
        }, status.HTTP_400_BAD_REQUEST

    try:
        graph = Graph().parse(data=request.data, format="ttl")
    except BadSyntax as e:
        return {"message": f"data is unreadable: {e}"}, status.HTTP_400_BAD_REQUEST

    if request.method == "PUT":
        model.graph.remove((None, None, None))

    model.add_graph(graph)

    current_app.building_motif.session.commit()

    return model.graph.serialize(format="ttl")


@blueprint.route("/<models_id>/validate", methods=(["POST"]))
def validate_model(models_id: int) -> flask.Response:
    # get model
    try:
        model = Model.load(models_id)
    except ModelNotFound:
        return {"message": f"ID: {models_id}"}, status.HTTP_404_NOT_FOUND

    # we will read the shape collections from the input
    shape_collections = []

    # get shacl_engine from the query params, defaults to the current engine
    shacl_engine = request.args.get("shacl_engine", None)

    # no body provided -- default to model manifest
    if request.content_length is None:
        shape_collections = [model.get_manifest()]
    else:
        # get body
        if request.content_type != "application/json":
            return flask.Response(
                {"message": "request content type must be json"},
                status.HTTP_400_BAD_REQUEST,
            )
        try:
            body = request.json
        except Exception as e:
            return {"message": f"cannot read body {e}"}, status.HTTP_400_BAD_REQUEST

        if body is not None and not isinstance(body, dict):
            return {"message": "body is not dict"}, status.HTTP_400_BAD_REQUEST
        body = body if body is not None else {}
        nonexistent_libraries = []
        for library_id in body.get("library_ids", []):
            try:
                shape_collection = Library.load(library_id).get_shape_collection()
                shape_collections.append(shape_collection)
            except LibraryNotFound:
                nonexistent_libraries.append(library_id)
        if len(nonexistent_libraries) > 0:
            return {
                "message": f"Libraries with ids {nonexistent_libraries} do not exist"
            }, status.HTTP_400_BAD_REQUEST

    # if shape_collections is empty, model.validate will default to the model's manifest
    vaildation_context = model.validate(
        shape_collections, error_on_missing_imports=False, shacl_engine=shacl_engine
    )

    return {
        "message": vaildation_context.report_string,
        "valid": vaildation_context.valid,
        "reasons": {
            focus_node: list(set(gd.reason() for gd in grahdiffs))
            for focus_node, grahdiffs in vaildation_context.diffset.items()
        },
    }, status.HTTP_200_OK


@blueprint.route("/<models_id>/validate_shape", methods=(["POST"]))
def validate_shape(models_id: int) -> flask.Response:
    # get model
    try:
        model = Model.load(models_id)
    except ModelNotFound:
        return {"message": f"ID: {models_id}"}, status.HTTP_404_NOT_FOUND

    # get body
    if request.content_type != "application/json":
        return flask.Response(
            {"message": "request content type must be json"},
            status.HTTP_400_BAD_REQUEST,
        )
    try:
        body = request.json
    except Exception as e:
        return {"message": f"cannot read body {e}"}, status.HTTP_400_BAD_REQUEST

    if not isinstance(body, dict):
        return {"message": "body is not dict"}, status.HTTP_400_BAD_REQUEST

    # shape collections
    shape_collections = []
    nonexistent_shape_collections = []
    for shape_collection_id in body.get("shape_collection_ids", []):
        try:
            shape_collection = ShapeCollection.load(shape_collection_id)
            shape_collections.append(shape_collection)
        except ShapeCollectionNotFound:
            nonexistent_shape_collections.append(shape_collection_id)
    if len(nonexistent_shape_collections) > 0:
        return {
            "message": f"shape collections with ids {nonexistent_shape_collections} do not exist"
        }, status.HTTP_400_BAD_REQUEST

    if body.get("target_class", None) is None:
        return {
            "message": "target class is required to execute this endpoint"
        }, status.HTTP_400_BAD_REQUEST

    shape_uris = [URIRef(shape_uri) for shape_uri in body.get("shape_uris", [])]
    target_class = URIRef(body.get("target_class"))

    # test
    compiled = model.compile(shape_collections)
    conformance = compiled.validate_model_against_shapes(
        shapes_to_test=shape_uris,
        target_class=target_class,
    )

    result = {}
    for shape_uri, validation_context in conformance.items():
        diffsets = validation_context.diffset.values()
        reasons = [diff.reason() for diffset in diffsets for diff in diffset]
        result[shape_uri] = reasons

    return result, status.HTTP_200_OK



from typing import Dict, List

import flask
from flask import Blueprint, current_app, jsonify
from flask_api import status
from rdflib import URIRef

from buildingmotif.api.serializers.library import serialize
from buildingmotif.database.errors import LibraryNotFound
from buildingmotif.dataclasses.shape_collection import ShapeCollection

blueprint = Blueprint("libraries", __name__)

get_shape_query = """
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    SELECT ?shape ?label ?desc
    WHERE {
        ?shape a sh:NodeShape ;
            sh:targetClass|sh:targetSubjectOf|sh:targetObjectsOf|sh:targetNode ?target .
        OPTIONAL { ?shape rdfs:label ?label }
        OPTIONAL { ?shape skos:description ?desc }
    }
"""


@blueprint.route("", methods=(["GET"]))
def get_all_libraries() -> flask.Response:
    """Get all libraries.

    :return: all libraries
    :rtype: flask.Response
    """
    db_libs = current_app.building_motif.table_connection.get_all_db_libraries()

    return jsonify(serialize(db_libs)), status.HTTP_200_OK


@blueprint.route("/shapes", methods=(["GET"]))
def get_all_shapes() -> flask.Response:
    """Get all shapes.

    :return: all shapes
    :rtype: flask.Response
    """
    definition_types = [
        URIRef("https://nrel.gov/BuildingMOTIF#Sequence_Of_Operations"),
        URIRef("https://nrel.gov/BuildingMOTIF#Analytics_Application"),
        URIRef("https://nrel.gov/BuildingMOTIF#System_Specification"),
    ]
    results: Dict[URIRef, List[Dict]] = {dt: [] for dt in definition_types}

    db_libs = current_app.building_motif.table_connection.get_all_db_libraries()
    for db_lib in db_libs:
        shape_collection = ShapeCollection.load(db_lib.shape_collection.id)
        for dt in definition_types:
            results[dt] += [
                {
                    "shape_uri": str(shape),
                    "label": label,
                    "library_name": db_lib.name,
                    "shape_collection_id": shape_collection.id,
                }
                for (shape, label) in shape_collection.get_shapes_of_definition_type(
                    dt, include_labels=True
                )
            ]

    return jsonify(results), status.HTTP_200_OK


@blueprint.route("/<library_id>", methods=(["GET"]))
def get_library(library_id: int) -> flask.Response:
    """Get library by id.

    :param library_id: library id
    :type library_id: int
    :return: requested library
    :rtype: flask.Response
    """
    try:
        db_lib = current_app.building_motif.table_connection.get_db_library(library_id)
    except LibraryNotFound:
        return {"message": f"ID: {library_id}"}, status.HTTP_404_NOT_FOUND

    return jsonify(serialize(db_lib)), status.HTTP_200_OK
